"""
This example calculates the magnetic field produced by a rectangular current 
loop inside a cylindrical magnetic shield.
"""

from mayavi import mlab

import matplotlib.pyplot as plt
import numpy as np

from bfieldtools.line_magnetics import magnetic_field
from bfieldtools.line_magnetics import scalar_potential
from bfieldtools.mesh_conductor import MeshConductor
from bfieldtools.mesh_conductor import StreamFunction
from bfieldtools.utils import load_example_mesh


def create_3d_grid(xx, yy, zz):
    """Creates a direct product grid from three 1D arrays (xx, yy and zz)
    that is appropriately formated for `scalar_potential` and `magnetic_field`.
    """
    X, Y, Z = np.meshgrid(xx, yy, zz, indexing="ij")

    x = X.ravel()
    y = Y.ravel()
    z = Z.ravel()

    return np.array([x, y, z]).T


# Defines a rectangular current loop offset by y from the xz plane
length = 0.3  # (m)
width = 6e-2  # (m)
ly = 6e-2  # (m)
loop_points = np.array(
    [
        [width / 2, ly, length / 2],
        [width / 2, ly, -length / 2],
        [-width / 2, ly, -length / 2],
        [-width / 2, ly, length / 2],
        [width / 2, ly, length / 2],
    ]
)


# Loads the cylinder shield geometry from examples
shield_mesh = load_example_mesh("closed_cylinder_remeshed")

# Shrinks the shield and rotates it by 90 degrees around the y axis.
shield_mesh.apply_scale(0.17)
shield_mesh.apply_transform([[0, 0, 1, 0], [0, 1, 0, 0], [1, 0, 0, 0], [0, 0, 0, 0]])

shield = MeshConductor(
    mesh_obj=shield_mesh, process=True, fix_normals=True, basis_name="vertex"
)

# Plots the complete geometry.
f1 = mlab.figure(None, bgcolor=(1, 1, 1), fgcolor=(0.5, 0.5, 0.5), size=(300, 300))
mlab.view(roll=45, azimuth=50, elevation=60, figure=f1)
mlab.plot3d(*loop_points.T, tube_radius=1e-3, figure=f1)
mlab.triangular_mesh(
    *shield_mesh.vertices.T,
    shield_mesh.faces,
    representation="wireframe",
    figure=f1,
    color=(0, 0, 0),
    opacity=0.05
)


# Calculates the stream function on the surface of the shield
shield = MeshConductor(mesh_obj=shield_mesh, process=True, fix_normals=True)

# In the following, we want to calculate scalar potentials on the surface of
# the shield. Since the potential created by the surface shield currents is
# discontinuous on the shield, we select the inner surface by taking a small
# offset along the inner normal.
d = 1e-5
shield_inner_points = shield_mesh.vertices - d * shield_mesh.vertex_normals

# Calculates the coupling matrix that relates the stream function of currents
# of the surface of the shield to scalar potential.
U_cpl_ssurf = shield.U_coupling(shield_inner_points)

# Takes the scalar potential created by the current loop at the shield and finds
# the stream function of surface currents that is necessary to compensate it.
U_loop_ssurf = scalar_potential(loop_points, shield_inner_points)
I_shield = np.linalg.solve(-U_cpl_ssurf, U_loop_ssurf)

# Represent the stream function using a dedicated class for visualization.
s_shield = StreamFunction(I_shield, shield)

# Visualizes the stream function on the shield.
f3 = mlab.figure(None, bgcolor=(1, 1, 1), fgcolor=(0.5, 0.5, 0.5), size=(300, 300))
mlab.view(roll=45, azimuth=45, elevation=60, figure=f3)
s_shield.plot(False, 64, figure=f3)
mlab.colorbar()
mlab.axes()

#%%
# The rest of the code calculates magnetic field on a 2D grid and displays
# the contributions of the current loop and the shield.


# Defines a grid in the yz plane.
x0 = 0
ny = 50
nz = 100

xx = np.linspace(x0, x0, 1)
yy = np.linspace(-0.9e-1, 0.9e-1, ny)
zz = np.linspace(-0.34 / 2, 0.34 / 2, nz)

grid = create_3d_grid(xx, yy, zz)


# Finds the scalar potential produced by the loop in the absence of shield.
U_loop1 = scalar_potential(loop_points, grid)
U_loop_pl1 = U_loop1.reshape(ny, nz)  # Reshapes for plotting.

# Finds the potential created by the shield.
U_cpl_shield1 = shield.U_coupling(grid)
U_shield1 = U_cpl_shield1 @ I_shield
U_shield_pl1 = U_shield1.reshape(ny, nz)  # Reshapes for plotting.

#%% Plots the scalar potential.

fig, axs = plt.subplots(2, 3)

plt.sca(axs[0, 0])
vmax = np.max(np.abs(U_loop1))
plt.contourf(zz, yy, U_loop_pl1, cmap=plt.get_cmap("RdBu"), vmax=vmax, vmin=-vmax)
plt.title("U: Loop only")
plt.xlabel("z (m)")
plt.ylabel("y (m)")
plt.colorbar()

plt.sca(axs[0, 1])
vmax = np.max(np.abs(U_shield1))
plt.contourf(zz, yy, U_shield_pl1, cmap=plt.get_cmap("RdBu"), vmax=vmax, vmin=-vmax)
plt.title("U: Shield only")
plt.xlabel("z (m)")
plt.ylabel("y (m)")
plt.colorbar()

plt.sca(axs[0, 2])
vmax = np.max(np.abs(U_loop1 + U_shield1))
plt.contourf(
    zz, yy, U_shield_pl1 + U_loop_pl1, cmap=plt.get_cmap("RdBu"), vmax=vmax, vmin=-vmax
)
plt.title("U: Total (loop+shield)")
plt.xlabel("z (m)")
plt.ylabel("y (m)")
plt.colorbar()


# Finds the magnetic field created by the loop in the absence of shield.
B_loop2 = magnetic_field(loop_points, grid)
B_loop_pl2 = B_loop2.reshape(ny, nz, 3)  # Reshapes for plotting.

# Finds the field created by the shield.
B_cpl_shield2 = shield.B_coupling(grid)
B_shield2 = B_cpl_shield2 @ I_shield
B_shield_pl2 = B_shield2.reshape(ny, nz, 3)  # Reshapes for plotting.

# Plots the magnetic field.
plt.sca(axs[1, 0])
vmax = np.max(np.abs(B_loop_pl2[:, :, 1])) * 1e4
plt.contourf(
    zz, yy, B_loop_pl2[:, :, 1] * 1e4, cmap=plt.get_cmap("RdBu"), vmax=vmax, vmin=-vmax
)
plt.title("B_y (Gauss @ 1amp), loop only")
plt.xlabel("z (m)")
plt.ylabel("x (m)")
plt.colorbar()

plt.sca(axs[1, 1])
vmax = np.max(np.abs(B_shield_pl2[:, :, 1])) * 1e4
plt.contourf(
    zz,
    yy,
    B_shield_pl2[:, :, 1] * 1e4,
    cmap=plt.get_cmap("RdBu"),
    vmax=vmax,
    vmin=-vmax,
)
plt.title("B_y (Gauss @ 1amp), shield only")
plt.xlabel("z (m)")
plt.ylabel("x (m)")
plt.colorbar()

plt.sca(axs[1, 2])
vmax = np.max(np.abs(B_loop_pl2[:, :, 1])) * 1e4
plt.contourf(
    zz,
    yy,
    (B_loop_pl2[:, :, 1] + B_shield_pl2[:, :, 1]) * 1e4,
    cmap=plt.get_cmap("RdBu"),
    vmax=vmax,
    vmin=-vmax,
)
plt.title("B_y (Gauss @ 1amp), loop+shield")
plt.xlabel("z (m)")
plt.ylabel("x (m)")
plt.colorbar()



import numpy as np

from bfieldtools import thermal_noise
from bfieldtools.utils import load_example_mesh
from bfieldtools.mesh_magnetics import magnetic_field_coupling

import trimesh
import pytest

from numpy.testing import (
    assert_array_almost_equal,
    assert_array_equal,
    assert_allclose,
    assert_equal,
)

from .test_mesh_conductor import _fake_mesh_conductor


def test_compute_modes():

    obj = _fake_mesh_conductor("unit_disc")

    DC_n = thermal_noise.compute_current_modes(
        obj, T=300, resistivity=1e-8, thickness=1e-3, mode="DC"
    )

    AC_n = thermal_noise.compute_current_modes(
        obj, freqs=np.array((0,)), T=300, resistivity=1e-8, thickness=1e-3, mode="AC"
    )

    points = np.array([[0.1, 0, 0.2], [-0.1, 0.1, -0.2]])

    obj.set_basis("vertex")

    DC_Bn_covar = thermal_noise.noise_covar(obj.B_coupling(points), DC_n)
    AC_Bn_covar = thermal_noise.noise_covar(obj.B_coupling(points), AC_n)

    assert_allclose(DC_Bn_covar, AC_Bn_covar[:, :, :, 0])

    DC_Bn_covar_dir = thermal_noise.noise_covar_dir(obj.B_coupling(points), DC_n)
    AC_Bn_covar_dir = thermal_noise.noise_covar_dir(obj.B_coupling(points), AC_n)

    assert_allclose(DC_Bn_covar_dir, AC_Bn_covar_dir[:, :, :, 0])

    DC_Bn_var = thermal_noise.noise_var(obj.B_coupling(points), DC_n)
    AC_Bn_var = thermal_noise.noise_var(obj.B_coupling(points), AC_n)

    assert_allclose(DC_Bn_var, AC_Bn_var[:, :, 0])

    thermal_noise.visualize_current_modes(obj.mesh, DC_n, scale=0.5, Nmodes=3)


def test_vs_analytic():

    obj = _fake_mesh_conductor("unit_disc")
    mesh = obj.mesh

    mesh.vertices, mesh.faces = trimesh.remesh.subdivide(mesh.vertices, mesh.faces)

    # Fix the simulation parameters
    d = 100e-6  # thickness
    sigma = 3.7e7  # conductivity
    res = 1 / sigma  # resistivity
    T = 300  # temperature
    kB = 1.38064852e-23  # Boltz
    mu0 = 4 * np.pi * 1e-7  # permeability of freespace

    # Compute the AC-current modes and visualize them
    vl, u = thermal_noise.compute_current_modes(
        obj=mesh, T=T, resistivity=res, thickness=d, mode="AC", return_eigenvals=True
    )

    # Define field points on z axis
    Np = 10
    z = np.linspace(0.2, 1, Np)
    fp = np.array((np.zeros(z.shape), np.zeros(z.shape), z)).T

    B_coupling = magnetic_field_coupling(
        mesh, fp, analytic=True
    )  # field coupling matrix

    # Compute noise variance
    B = np.sqrt(thermal_noise.noise_var(B_coupling, vl))

    # Next, we compute the DC noise without reference to the inductance
    vl_dc, u_dc = thermal_noise.compute_current_modes(
        obj=mesh, T=T, resistivity=res, thickness=d, mode="DC", return_eigenvals=True
    )

    # Compute noise variance
    B_dc = np.sqrt(thermal_noise.noise_var(B_coupling, vl_dc))

    # Calculate Bz noise using analytical formula and plot the results
    r = 1
    Ban = (
        mu0
        * np.sqrt(sigma * d * kB * T / (8 * np.pi * z**2))
        * (1 / (1 + z**2 / r**2))
    )

    assert_allclose(B_dc[:, 2], B[:, 2, 0])

    assert_allclose(Ban, B[:, 2, 0], rtol=5e-2)



"""
Linear dipole density on a triangle
=========================================

Test and validation of potential of linearly distributed dipolar density

For the math see:
        J. C. de Munck, "A linear discretization of the volume mesh_conductor
        boundary integral equation using analytically integrated elements
        (electrophysiology application),"
        in IEEE Transactions on Biomedical Engineering,
        vol. 39, no. 9, pp. 986-990, Sept. 1992.
        doi: 10.1109/10.256433


"""

import numpy as np
import matplotlib.pyplot as plt
import sys

path = "/m/home/home8/80/makinea1/unix/pythonstuff/bfieldtools"
if path not in sys.path:
    sys.path.insert(0, path)

from bfieldtools.integrals import triangle_potential_dipole_linear
from bfieldtools.integrals import omega
from bfieldtools.utils import tri_normals_and_areas

#%%
#%% Test potential shape slightly above the surface
# points = np.array([[0,0,0],
#                   [1,0,0],
#                   [0,1,0]])
#
# tris = np.array([[0,1,2]])
# p_tris = points[tris]

points = np.array([[0, 0, 0], [1, 1, 0], [1, -1, 0], [-1, -1, 0], [-1, 1, 0]])

tris = np.array([[0, 1, 2], [0, 2, 3], [0, 3, 4], [0, 4, 1]])
tris = np.flip(tris, axis=-1)
p_tris = points[tris]

# Evaluation points
Nx = 100
xx = np.linspace(-2, 2, Nx)
X, Y = np.meshgrid(xx, xx, indexing="ij")
Z = np.zeros_like(X) + 0.01
p_eval = np.array([X, Y, Z]).reshape(3, -1).T

# Difference vectors
RR = p_eval[:, None, None, :] - p_tris[None, :, :, :]
tn, ta = tri_normals_and_areas(points, tris)

pot = triangle_potential_dipole_linear(RR, tn, ta)

# Plot shapes
f, ax = plt.subplots(1, 3)
for i in range(3):
    plt.sca(ax[i])
    plt.imshow(
        pot[:, 2, i].reshape(Nx, Nx), extent=(xx.min(), xx.max(), xx.max(), xx.min())
    )
    plt.colorbar(orientation="horizontal")
    if i == 0:
        plt.ylabel("x")
        plt.xlabel("y")

#%%
#%% Test summation formula
# NOTE: the sign of tilde(omega)_i in the bfieldtools
# (triangle_potential_dipole_linear) is equal to -omega_i
# in the de Munck's paper refered above
pot_sum = triangle_potential_dipole_linear(RR, tn, ta).sum(axis=-1)
solid_angle = omega(RR)

# Plot shapes
f, ax = plt.subplots(1, 3)
plt.sca(ax[0])
plt.title("Sum of potentials")
plt.imshow(pot_sum[:, 0].reshape(Nx, Nx), vmin=0, vmax=pot_sum.max())
plt.colorbar(orientation="horizontal")
plt.sca(ax[1])
plt.title("Negative solid angle")
plt.imshow(-solid_angle[:, 0].reshape(Nx, Nx), vmin=0, vmax=pot_sum.max())
plt.colorbar(orientation="horizontal")
plt.sca(ax[2])
plt.title("Abs difference")
plt.imshow(
    abs((-solid_angle[:, 0] - pot_sum[:, 0])).reshape(Nx, Nx),
    vmin=0,
    vmax=pot_sum.max() / 1e16,
)
plt.colorbar(orientation="horizontal", pad=-0.2)
plt.axis("image")

plt.tight_layout()


#%%
#%% Test asymptotic behavour
def dip_potential(Reval, Rdip, moment):
    R = Reval - Rdip
    r = np.linalg.norm(R, axis=1)
    return (moment * R).sum(axis=1) / r**3


# Center of mass
Rdip = points.mean(axis=0)
# Moment
m = ta[0] * tn[0]
# Eval points
Neval = 100
p_eval2 = np.zeros((Neval, 3))
z = np.linspace(0.01, 100, Neval)
p_eval2[:, 2] = z
p_eval2 += Rdip


plt.figure()

# Plot dipole field approximating uniform dipolar density
plt.semilogy(z, dip_potential(p_eval2, Rdip, m))
# Plot sum of the linear dipoles
RR = p_eval2[:, None, None, :] - p_tris[None, :, :, :]
pot = triangle_potential_dipole_linear(RR, tn, ta)
plt.semilogy(z, pot.sum(axis=-1)[:, 0])



"""
Example of spherical harmonics tools and visualization
============================================================
"""

import numpy as np
from mayavi import mlab
import matplotlib.pyplot as plt

from bfieldtools.sphtools import SphBasis
from bfieldtools import sphtools

# Create basis object for plotting, evaluation of inner products etc.
sph = SphBasis(40)

#%%
# Plot Spherical harmonics Ylms on the sphere
# Plot Ylms up to l = 4
mlab.figure(bgcolor=(1, 1, 1))
obj = sphtools.plotYlms(sph, lmax=4)

# Plot Ylm with l=3, m=3
mlab.figure(bgcolor=(1, 1, 1))
obj = sphtools.plotYlm(sph, l=3, m=3)

#%%
# Plot vector spherical harmonics Vlms / Wlms / Xlms on the sphere

# Plot Vlm with l=2, m=2
mlab.figure(bgcolor=(1, 1, 1))
obj = sphtools.plotVlm(sph, l=2, m=2)
obj.scene.z_plus_view()

# Plot Wlm with the same order and degree, l=2, m=2
mlab.figure(bgcolor=(1, 1, 1))
obj = sphtools.plotWlm(sph, l=2, m=2)
obj.scene.z_plus_view()

# Plot Xlm with the same order and degree, l=2, m=2
mlab.figure(bgcolor=(1, 1, 1))
obj = sphtools.plotXlm(sph, l=2, m=2)
obj.scene.z_plus_view()


#%%
# Plot fields of vector spherical harmonics Vlms / Wlms on a volume

# Plot BVlm with l=2, m=2
offset = np.array((0, 0, 3))  # add offset due to divergence in origin
mlab.figure(bgcolor=(1, 1, 1))
obj = sphtools.plotBVlm_volume(sph, l=2, m=2, lim=1, Np=10, offset=offset)

# Plot BWlm with l=2, m=2
offset = np.array((0, 0, 0))  # include origin
mlab.figure(bgcolor=(1, 1, 1))
obj = sphtools.plotBWlm_volume(sph, l=2, m=2, lim=1, Np=10, offset=offset)

#%%
# Test inner products in quadrature points defined in sph-object

Ylm1 = sphtools.ylm(2, 1, sph.sqp[:, 1], sph.sqp[:, 2])
Ylm2 = sphtools.ylm(2, 0, sph.sqp[:, 1], sph.sqp[:, 2])
print(
    "Inner product between Ylm_20 and Ylm_21 is %f"
    % sph.innerproduct(Ylm1[:, None], Ylm2[:, None])
)
print(
    "Inner product between Ylm_20 and Ylm_20 is %f"
    % sph.innerproduct(Ylm2[:, None], Ylm2[:, None])
)

Vlm1 = sphtools.Vlm(1, 0, sph.sqp[:, 1], sph.sqp[:, 2])
Vlm2 = sphtools.Vlm(2, 0, sph.sqp[:, 1], sph.sqp[:, 2])
print("Inner product between V_10 and V_20 is %f" % sph.innerproduct(Vlm1, Vlm2))
print("Inner product between V_10 and V_10 is %f" % sph.innerproduct(Vlm1, Vlm1))

Wlm1 = sphtools.Wlm(1, 0, sph.sqp[:, 1], sph.sqp[:, 2])
Wlm2 = sphtools.Wlm(2, 0, sph.sqp[:, 1], sph.sqp[:, 2])
print("Inner product between W_10 and W_20 is %f" % sph.innerproduct(Wlm1, Wlm2))
print("Inner product between W_10 and W_10 is %f" % sph.innerproduct(Wlm1, Wlm1))

Xlm1 = sphtools.Xlm(1, 0, sph.sqp[:, 1], sph.sqp[:, 2])
Xlm2 = sphtools.Xlm(2, 0, sph.sqp[:, 1], sph.sqp[:, 2])
print("Inner product between X_10 and X_20 is %f" % sph.innerproduct(Xlm1, Xlm2))
print("Inner product between X_10 and X_10 is %f" % sph.innerproduct(Xlm1, Xlm1))

print("Inner product between W_10 and V_10 is %f" % sph.innerproduct(Wlm1, Vlm1))
print("Inner product between X_10 and V_10 is %f" % sph.innerproduct(Xlm1, Vlm1))
print("Inner product between X_10 and W_10 is %f" % sph.innerproduct(Xlm1, Vlm1))


#%%
# Simple example of the calculation of sph spectrum of magnetic field
# defined on the sphere using the inner product.

# define homogeneous magnetic field along z-axis
B = np.zeros(sph.sqp.shape)
B[:, 2] = sph.qp.points[:, 0] / np.max(sph.qp.points[:, 0])

# add noise
B += 0.1 * np.random.randn(B.shape[0], B.shape[1])

B = sphtools.cartvec2sph(sph.sqp, B)  # transform to spherical coordinates

coeffs = sph.avsphspectra(B, 7)  # calculate spectra of the r**l terms

# Plot the spectra
plt.figure()
plt.semilogy(coeffs**2)



"""
Biplanar coil design
====================

Example showing a basic biplanar coil producing homogeneous field in a target
region between the two coil planes.

"""

import numpy as np
import matplotlib.pyplot as plt
from mayavi import mlab
import trimesh


from bfieldtools.mesh_conductor import MeshConductor
from bfieldtools.coil_optimize import optimize_streamfunctions
from bfieldtools.viz import plot_cross_section
from bfieldtools.utils import combine_meshes, load_example_mesh


# Load simple plane mesh that is centered on the origin
planemesh = load_example_mesh("10x10_plane_hires")

# Specify coil plane geometry
center_offset = np.array([0, 0, 0])
standoff = np.array([0, 3, 0])

# Create coil plane pairs
coil_plus = trimesh.Trimesh(
    planemesh.vertices + center_offset + standoff, planemesh.faces, process=False
)

coil_minus = trimesh.Trimesh(
    planemesh.vertices + center_offset - standoff, planemesh.faces, process=False
)

joined_planes = combine_meshes((coil_plus, coil_minus))

# Create mesh class object
coil = MeshConductor(
    mesh_obj=joined_planes, fix_normals=True, basis_name="suh", N_suh=100
)

#%%
# Set up target and stray field points

# Here, the target points are on a volumetric grid within a sphere

center = np.array([0, 0, 0])

sidelength = 1.5
n = 8
xx = np.linspace(-sidelength / 2, sidelength / 2, n)
yy = np.linspace(-sidelength / 2, sidelength / 2, n)
zz = np.linspace(-sidelength / 2, sidelength / 2, n)
X, Y, Z = np.meshgrid(xx, yy, zz, indexing="ij")

x = X.ravel()
y = Y.ravel()
z = Z.ravel()

target_points = np.array([x, y, z]).T

# Turn cube into sphere by rejecting points "in the corners"
target_points = (
    target_points[np.linalg.norm(target_points, axis=1) < sidelength / 2] + center
)


# Here, the stray field points are on a spherical surface
stray_radius = 20
stray_points_mesh = trimesh.creation.icosphere(subdivisions=3, radius=stray_radius)
stray_points = stray_points_mesh.vertices + center

n_stray_points = len(stray_points)


#%%
# Create bfield specifications used when optimizing the coil geometry

# The absolute target field amplitude is not of importance,
# and it is scaled to match the C matrix in the optimization function

target_field = np.zeros(target_points.shape)
target_field[:, 0] += 1

target_spec = {
    "coupling": coil.B_coupling(target_points),
    "abs_error": 0.01,
    "target": target_field,
}
stray_spec = {
    "coupling": coil.B_coupling(stray_points),
    "abs_error": 0.01,
    "target": np.zeros((n_stray_points, 3)),
}

bfield_specification = [target_spec, stray_spec]

#%%
## Compute the optimal stream function, either using a numerical solver or regularized least squares

import mosek

coil.s, prob = optimize_streamfunctions(
    coil,
    [target_spec, stray_spec],
    objective="minimum_ohmic_power",
    solver="MOSEK",
    solver_opts={"mosek_params": {mosek.iparam.num_threads: 8}},
)


#%%
# Plot the optimized stream function, then discretize it and plot coil windings and the resultant magnetic field

coil.s.plot()

loops = coil.s.discretize(N_contours=10)

loops.plot_loops()

B_target = loops.magnetic_field(target_points)
mlab.quiver3d(*target_points.T, *B_target.T)


#%%
# Lets also do the same coil optimization using regularized least-squares.
# Now we can't specify inequality constraints (e.g. use error margins in the specification).


from bfieldtools.coil_optimize import optimize_lsq

coil.s2 = optimize_lsq(
    coil, [target_spec, stray_spec], objective="minimum_ohmic_power", reg=1e6
)


#%%
# Plot the optimized stream function, then discretize it and plot coil windings and the resultant magnetic field

coil.s2.plot()

loops2 = coil.s2.discretize(N_contours=10)

loops2.plot_loops()

B_target = loops2.magnetic_field(target_points)
mlab.quiver3d(*target_points.T, *B_target.T)


#%%
# Plot cross-section of magnetic field and magnetic potential of the discretized loops

x = y = np.linspace(-12, 12, 250)
X, Y = np.meshgrid(x, y, indexing="ij")


points = np.zeros((X.flatten().shape[0], 3))
points[:, 0] = X.flatten()
points[:, 1] = Y.flatten()

B = loops2.magnetic_field(points)
U = loops2.scalar_potential(points)

U = U.reshape(x.shape[0], y.shape[0])
B = B.T[:2].reshape(2, x.shape[0], y.shape[0])

lw = np.sqrt(B[0] ** 2 + B[1] ** 2)

lw = 2 * lw / np.max(lw)

plot_cross_section(X, Y, U, log=False, contours=False)

seed_points = points[:, :2] * 0.3

plt.streamplot(
    x,
    y,
    B[0],
    B[1],
    density=2,
    linewidth=lw,
    color="k",
    integration_direction="both",
    start_points=seed_points,
)
plt.axis("equal")
plt.axis("off")

plt.plot([-5, 5], [-3, -3], "k", linewidth=3, alpha=1)
plt.plot([-5, 5], [3, 3], "k", linewidth=3, alpha=1)

plt.tight_layout()



"""
Analytic integral for vectorized field / potential computation

"""

import numpy as np


def determinant(a):
    det = a[..., 0, 0] * (a[..., 1, 1] * a[..., 2, 2] - a[..., 2, 1] * a[..., 1, 2])
    det += a[..., 0, 1] * (a[..., 1, 2] * a[..., 2, 0] - a[..., 2, 2] * a[..., 1, 0])
    det += a[..., 0, 2] * (a[..., 1, 0] * a[..., 2, 1] - a[..., 2, 0] * a[..., 1, 1])
    return det


def norm(vecs):
    return np.sqrt(np.einsum("...i,...i", vecs, vecs))


def cross(r1, r2):
    """Cross product without overhead"""
    result = np.zeros(r1.shape)
    result[..., 0] = r1[..., 1] * r2[..., 2] - r1[..., 2] * r2[..., 1]
    result[..., 1] = r1[..., 2] * r2[..., 0] - r1[..., 0] * r2[..., 2]
    result[..., 2] = r1[..., 0] * r2[..., 1] - r1[..., 1] * r2[..., 0]
    return result


def gamma0(R, reg=1e-13, symmetrize=True):
    """Integrals over the edges of a triangle called gamma_0 (line charge potentials).

    **NOTE: MAY NOT BE VERY PRECISE FOR POINTS DIRECTLY AT TRIANGLE
    EDGES.**

    Parameters
    ----------
    R : (N, 3, 3) array of points (Neval, Nverts, xyz)

    Returns
    -------
    res: array (Neval, Nverts)
        The analytic integrals for each vertex/edge

    """
    edges = np.roll(R[0], 1, -2) - np.roll(R[0], 2, -2)
    #    dotprods1 = np.sum(np.roll(R, 1, -2)*edges, axis=-1)
    #    dotprods2 = np.sum(np.roll(R, 2, -2)*edges, axis=-1)
    dotprods1 = np.einsum("...i,...i", np.roll(R, 1, -2), edges)
    dotprods2 = np.einsum("...i,...i", np.roll(R, 2, -2), edges)
    en = norm(edges)
    del edges
    n = norm(R)
    # Regularize s.t. neither the denominator or the numerator can be zero
    # Avoid numerical issues directly at the edge
    nn1 = np.roll(n, 2, -1) * en
    nn2 = np.roll(n, 1, -1) * en
    res = np.log((nn1 + dotprods2 + reg) / (nn2 + dotprods1 + reg))

    # Symmetrize the result since on the negative extension of the edge
    # there's division of two small values resulting numerical instabilities
    # (also incompatible with adding the reg value)
    if symmetrize:
        res2 = -np.log((nn1 - dotprods2 + reg) / (nn2 - dotprods1 + reg))
        res = np.where(dotprods1 + dotprods2 > 0, res, res2)
    res /= en
    return -res  # TODO: there should be minus, since we want this to be positive


def omega(R):
    """Calculate the solid angle of a triangles

    see
    A. Van Oosterom and J. Strackee
    IEEE TRANSACTIONS ON BIOMEDICAL ENGINEERING,
    VOL. BME-30, NO. 2, 1983

    Parameters
    ----------
    R : array of points (Neval, (Ntri), Nverts, xyz)
        Points correspond to relative coordinates (x,y,z) of
        N triangles/evaluation points for
        the 3 corners of the triangles/triangle.

        Neval can be number of evaluation points for the same triangle
        or number of triangles for the same evaluation points

    Returns
    -------
    sa: (Neval, (Ntri))
        Solid angles of triangle(s) at evaluation points
    """
    # Distances
    d = norm(R)
    # Scalar triple products
    stp = determinant(R)
    # Denominator
    denom = np.prod(d, axis=-1)
    for i in range(3):
        j = (i + 1) % 3
        k = (i + 2) % 3
        #        denom += np.sum(R[..., i, :]*R[..., j, :], axis=-1)*d[..., k]
        denom += np.einsum("...i,...i,...", R[..., i, :], R[..., j, :], d[..., k])
    # Solid angles
    sa = 2 * np.arctan2(stp, denom)
    return sa


def triangle_potential_uniform(R, tn, planar=False):
    """1/r potential of a uniform triangle

    see
    A. S. Ferguson, Xu Zhang and G. Stroink,
    "A complete linear discretization for calculating the magnetic field
    using the boundary element method,"
    in IEEE Transactions on Biomedical Engineering,
    vol. 41, no. 5, pp. 455-460, May 1994.
    doi: 10.1109/10.293220

    Parameters
    ----------

    R : (Neval, (Ntri), 3, 3) array
        Displacement vectors (Neval, (Ntri), Ntri_verts, xyz)
    tn : ((Ntri), 3) array
        Triangle normals (Ntri, dir)
    planar: boolean
        If True, use planar geometry assumption for speed

    Returns
    -------
    result: result:  ndarray (Neval, (Ntri))
        Resultant 1/r potential for each triangle (Ntri)
        at the field evaluation points (Neval)

    """
    if len(R.shape) > 3:
        tn_ax = tn[:, None, :]
    else:
        tn_ax = tn
    summands = np.sum(
        tn_ax * np.cross(np.roll(R, 1, -2), np.roll(R, 2, -2), axis=-1), axis=-1
    )
    #    summands = -gamma0(R)*np.sum(tn_ax*np.cross(np.roll(R, 1, -2),
    #                                               np.roll(R, 2, -2), axis=-1), axis=-1)
    result = np.einsum("...i,...i", -gamma0(R), summands)
    if not planar:
        #        csigned = np.sum(np.take(R, 0, -2)*tn, axis=-1)
        csigned = np.einsum("...i,...i", np.take(R, 0, -2), tn)
        result -= csigned * omega(R)
    else:
        print("Assuming all the triangles are in the same plane!")
    #        result = np.sum(summands, axis=-1)
    return result


def triangle_potential_approx(R, ta, reg=1e-12):
    """1/r potential of a uniform triangle using centroid approximation

    Calculates 1/R potentials for triangle centroids
    (The singularity at the centroid is handled with the very small
    reg value, but anyway the values close to the centroid are inexact)

    Parameters
    ----------
    R : (N, (Ntri), 3, 3) array
        Displacement vectors (Neval, ...., Ntri_verts, xyz)
    ta : (Ntri) array
        Triangle areas

    reg: float
        Regularization value used in approximation

    Returns
    -------
    result: result:  ndarray (...., Ntri, Ntri_verts)
        Resultant 1/r potential for each node (Ntri_verts)
        in each triangle (Ntri) in the displacement vectors R

    """
    result = 1 / (norm(np.mean(R, axis=-2)) + reg) * ta
    return result


def potential_dipoles(R, face_normals, face_areas):
    """Approximate the potential of linearly varying dipole density by
        by dipoles at each face

    Parameters
            R : (Neval, Ntri, 3, 3) array
                Displacement vectors (Neval, Ntri, Ntri_verts, xyz)
            face_normals: (Ntri, 3) normals of each fame
            face_areas

    Return
        Potential approximation for vertex in each face
        pot: (Neval, Ntri, Ntriverts)
    """
    nn = face_normals
    # Calculate quadrature points corresponding to linear shape functions (Ok?)
    weights = np.array([[0.5, 0.25, 0.25], [0.25, 0.5, 0.25], [0.25, 0.25, 0.5]])
    #    weights = np.eye(3)
    #    weights = np.ones((3,3))/3
    # Combine vertices for quadrature points
    Rquad = np.einsum("...ij,ik->...kj", R, weights)
    pot = np.einsum("ik, ...ijk->...ij", nn, Rquad) / (norm(Rquad) ** 3)
    pot = pot * (face_areas[:, None] / 3)

    return pot


def triangle_potential_dipole_linear(R, tn, ta, planar=False):
    """Potential of dipolar density with magnitude of a
    linear shape function on a triangle, "omega_i" in de Munck's paper

    see
    J. C. de Munck, "A linear discretization of the volume conductor
    boundary integral equation using analytically integrated elements
    (electrophysiology application),"
    in IEEE Transactions on Biomedical Engineering,
    vol. 39, no. 9, pp. 986-990, Sept. 1992.
    doi: 10.1109/10.256433

    Parameters
    ----------

    R : (..., Ntri, 3, 3) array
        Displacement vectors (...., Ntri, Ntri_verts, xyz)
    tn : ((Ntri), 3) array
        Triangle normals (Ntri, dir)
    ta : (Ntri), array
        Triangle areas (Ntri, dir)
    planar: boolean
        If True, use planar geometry assumption for speed

    Returns
    -------
    result:  ndarray (...., Ntri, Ntri_verts)
        Resultant dipolar potential for each shape functions (Ntri_verts)
        in each triangle (Ntri) at the points
        corresponding to displacement vectors in R

    """
    if len(R.shape) > 3:
        tn_ax = tn[:, None, :]
    else:
        tn_ax = tn
    # Volumes of tetrahedron between field evaluation point and the triangle
    #    det = np.sum(np.cross(np.roll(R, 2, -2),
    #                          np.roll(R, 1, -2), axis=-1)*R, axis=-1)
    det = determinant(R)
    # Edges opposite to the nodes
    edges = np.roll(R[0], 1, -2) - np.roll(R[0], 2, -2)
    # Latter part of omega_i integral in de Munck
    #    result = np.sum(np.sum(gamma0(R)[..., None]*edges, axis=-2)[...,None,:]*edges, axis=-1)
    result = np.einsum(
        "...i,...ik,...jk,...->...j",
        gamma0(R),
        edges,
        edges,
        det / (2 * ta),
        optimize=True,
    )
    #    result *= (det/(2*ta))[..., None] # TODO: IS DET SIGN OK?
    if not planar:
        # First part of the integral
        # Note: tn normalized version of n-vector in de Munck
        lin_coeffs = np.sum(
            tn_ax * cross(np.roll(R, 2, -2), np.roll(R, 1, -2)), axis=-1
        )
        result += lin_coeffs * omega(R)[..., :, None]
    else:
        print("Assuming all the triangles are in the same plane!")
    return result / (2 * ta[:, None])



#%%
import numpy as np
import matplotlib.pyplot as plt
import trimesh
from mayavi import mlab
from scipy.linalg import eigh

from bfieldtools.thermal_noise import (
    compute_current_modes,
    visualize_current_modes,
    noise_covar,
    noise_var,
)
from bfieldtools.mesh_magnetics import magnetic_field_coupling

from bfieldtools import utils

import pkg_resources

font = {"family": "normal", "weight": "normal", "size": 16}
plt.rc("font", **font)


#%%
# Fix the simulation parameters
d = 1e-3
sigma = 3.8e7
T = 293
kB = 1.38064852e-23
mu0 = 4 * np.pi * 1e-7

freqs = np.array((0,))

Np = 200

z = np.linspace(0.05, 10, Np)
fp = np.array((np.zeros(z.shape), np.zeros(z.shape), z)).T

Niters = 3
Nfaces = np.zeros(Niters)
Bn = np.zeros((Niters, Np))

modinds = np.array((0, 4, 9, 49, 99, 249))
Nmods = len(modinds)

Nchunks = 4
quad_degree = 4

for i in range(Niters):
    if i == 0:
        mesh = trimesh.load(
            pkg_resources.resource_filename(
                "bfieldtools", "example_meshes/disc_finer.stl"
            )
        )

    if i == 1:
        mesh = trimesh.load(
            pkg_resources.resource_filename(
                "bfieldtools", "example_meshes/disc_extrafine.stl"
            )
        )
    if i == 2:
        mesh = trimesh.load(
            pkg_resources.resource_filename(
                "bfieldtools", "example_meshes/disc_extremelyfine.stl"
            )
        )
    B_coupling = magnetic_field_coupling(mesh, fp, analytic=True)

    Nfaces[i] = mesh.faces.shape[0]

    vl, u = compute_current_modes(
        obj=mesh,
        T=T,
        resistivity=1 / sigma,
        thickness=d,
        mode="AC",
        return_eigenvals=True,
    )

    scene = mlab.figure(
        None, bgcolor=(1, 1, 1), fgcolor=(0.5, 0.5, 0.5), size=(800, 800)
    )

    visualize_current_modes(mesh, vl[:, :, 0], 20, 5, contours=True)

    for m in range(Nmods):
        scene = mlab.figure(
            figure=m, bgcolor=(1, 1, 1), fgcolor=(0.5, 0.5, 0.5), size=(800, 800)
        )

        s = mlab.triangular_mesh(
            *mesh.vertices.T,
            mesh.faces,
            scalars=vl[:, modinds[m], 0],
            colormap="bwr",
            line_width=12
        )
        limit = np.max(np.abs(vl[:, modinds[m], 0]))
        s.module_manager.scalar_lut_manager.data_range = np.array([-limit, limit])
        s.enable_contours = True
        surface = scene.children[0].children[0].children[0].children[0]
        surface.contour.number_of_contours = 10
        scene.scene.z_plus_view()
        scene.scene.camera.position = [
            -5.513350804725592e-05,
            -1.691800821806977e-05,
            4.020380431659883,
        ]
        scene.scene.camera.focal_point = [
            -5.513350804725592e-05,
            -1.691800821806977e-05,
            0.0,
        ]
        scene.scene.camera.view_angle = 30.0
        scene.scene.camera.view_up = [0.0, 1.0, 0.0]
        scene.scene.camera.clipping_range = [3.7658023470473827, 4.350539189462033]
        scene.scene.camera.compute_view_plane_normal()
        scene.scene.render()

    for m in range(Nmods):
        mlab.figure(m)

    mlab.close(all=True)

    scene = mlab.figure(
        None, bgcolor=(1, 1, 1), fgcolor=(0.5, 0.5, 0.5), size=(800, 800)
    )
    s = mlab.triangular_mesh(*mesh.vertices.T, mesh.faces)
    scene.scene.z_minus_view()
    surface = scene.children[0].children[0].children[0].children[0]
    surface.actor.property.representation = "wireframe"
    surface.actor.mapper.scalar_visibility = False
    scene.scene.camera.position = [0.0, 0.0, -4.515786458791532]
    scene.scene.camera.focal_point = [0.0, 0.0, 0.0]
    scene.scene.camera.view_angle = 30.0
    scene.scene.camera.view_up = [0.0, 1.0, 0.0]
    scene.scene.camera.clipping_range = [4.229838328573525, 4.886628590046963]
    scene.scene.camera.compute_view_plane_normal()
    scene.scene.render()

    Bvar = noise_var(B_coupling, vl)
    Bn[i] = np.sqrt(Bvar[:, 2, 0])

    if i == 0:
        u0 = u
    if i == 1:
        u1 = u
    if i == 2:
        u2 = u

    print(i)

figw = 3.75

plt.figure(figsize=(figw, 5))
plt.loglog(1 / u0 * 1e3, linewidth=2, label="N = %i" % Nfaces[0])
plt.loglog(1 / u1 * 1e3, linewidth=2, label="N = %i" % Nfaces[1])
plt.loglog(1 / u2 * 1e3, linewidth=2, label="N = %i" % Nfaces[2])
plt.ylim((0.1, 10))
plt.grid()
plt.gca().spines["right"].set_visible(False)
plt.gca().spines["top"].set_visible(False)
plt.xlabel("Mode index")
plt.ylabel("Time constant (ms)")
plt.legend(frameon=False, loc=3)
plt.tight_layout()


r = 1
Ban = (
    mu0
    * np.sqrt(sigma * d * kB * T / (8 * np.pi * z**2))
    * (1 / (1 + z**2 / r**2))
)

plt.figure(figsize=(figw, 5))
plt.loglog(z, Ban * 1e15, linewidth=2, label="Analytic")
plt.loglog(z, Bn[2] * 1e15, "x", label="N = %i" % Nfaces[2])
plt.grid()
plt.gca().spines["right"].set_visible(False)
plt.gca().spines["top"].set_visible(False)
plt.legend(frameon=False, loc=3)
plt.xlabel("Distance (z/R)")
plt.ylabel(r"$B_z$ noise at DC (fT/rHz)")
plt.tight_layout()

plt.figure(figsize=(figw, 5))
for i in range(Niters):
    plt.loglog(
        z,
        np.abs((Bn[i] - Ban)) / np.abs(Ban) * 100,
        linewidth=2,
        label="N = %i" % Nfaces[i],
    )
plt.grid()
plt.gca().spines["right"].set_visible(False)
plt.gca().spines["top"].set_visible(False)
plt.xlabel("Distance (z/R)")
plt.ylabel("Relative error (%)")
plt.legend(frameon=False)
plt.tight_layout()

N = 100
Nmodes = np.floor(np.linspace(1, vl.shape[1], N)).astype("int")
Bm = np.zeros((N, Np))

for k in range(N):
    Bvar = noise_var(B_coupling, vl, Nmodes=Nmodes[k])
    Bm[k] = np.sqrt(Bvar[:, 2, 0])


plt.figure(figsize=(figw, 5))
Ban0 = (
    mu0
    * np.sqrt(sigma * d * kB * T / (8 * np.pi * z[0] ** 2))
    * (1 / (1 + z[0] ** 2 / r**2))
)
plt.semilogy(
    Nmodes,
    np.abs((Bm[:, 0] - Ban0)) / np.abs(Ban0) * 100,
    linewidth=2,
    label="z/R = %.2f" % z[0],
)

Ban0 = (
    mu0
    * np.sqrt(sigma * d * kB * T / (8 * np.pi * z[1] ** 2))
    * (1 / (1 + z[1] ** 2 / r**2))
)
plt.semilogy(
    Nmodes,
    np.abs((Bm[:, 1] - Ban0)) / np.abs(Ban0) * 100,
    linewidth=2,
    label="z/R = %.2f" % z[1],
)

plt.grid()
plt.gca().spines["right"].set_visible(False)
plt.gca().spines["top"].set_visible(False)
plt.xlabel("Number of modes")
plt.ylabel("Relative error (%)")
plt.legend(frameon=False)
plt.tight_layout()



# -*- coding: utf-8 -*-
"""
Flatten mesh using conformal mapping
=============================================

Map 3D mesh to a 2D (complex) plane with angle-preserving (conformal) mapping

Based on these course notes
https://www.cs.cmu.edu/~kmcrane/Projects/DDG/
section 7.4.

"""

import numpy as np
from bfieldtools.flatten_mesh import flatten_mesh
from bfieldtools.flatten_mesh import mesh2plane
from bfieldtools.flatten_mesh import plane2mesh
from bfieldtools.utils import load_example_mesh

#%% Determine 2D parameterization and plot coordinate function on the 3D mesh
from mayavi import mlab
from bfieldtools.viz import plot_data_on_vertices, plot_mesh, plot_data_on_faces

mesh = load_example_mesh("meg_helmet", process=False)
u, v, mesh2d = flatten_mesh(mesh, _lambda=0.80)

plot_data_on_vertices(mesh, u, ncolors=15)
plot_data_on_vertices(mesh, v, ncolors=15)

#%% Determine lambda with smallest area distortion
# lls = np.linspace(0.01,1.0, 100)
# mm = []
# for ll in lls:
#     u, v, mesh2d = flatten_mesh(mesh, _lambda=ll)
#     d = mesh2d.area_faces / mesh.area_faces
#     mm.append(np.std(d)/np.mean(d))
#     print(np.std(d)/np.mean(d))
# plt.plot(lls, mm)

#%% Plot flattened mesh and area distortion on faces
plot_data_on_faces(mesh2d, mesh2d.area_faces / mesh.area_faces)

#%% Plot gradient of the two coordinate functions and the cosine of the angle between the gradients
from bfieldtools.mesh_calculus import gradient

gx = gradient(u, mesh)
gy = gradient(v, mesh)
cos = np.sum(gx * gy, axis=0) / (
    np.linalg.norm(gx, axis=0) * np.linalg.norm(gy, axis=0)
)
plot_data_on_faces(mesh, cos, vmin=-1, vmax=1)
mlab.quiver3d(*mesh.triangles_center.T, *gx, color=(1, 0, 0), mode="arrow")
q = mlab.quiver3d(*mesh.triangles_center.T, *gy, color=(0, 0, 1), mode="arrow")
q.scene.isometric_view()

#%% Map hexagonal grid from 2d to the 3D mesh
d = np.sqrt(3 / 4)
m = np.min((u.min(), v.min()))
mm = np.min((u.max(), v.max()))
xx = np.linspace(m * 1.05, mm * 1.05, 12)
yy = np.linspace(m * 1.05, mm * 1.05, 12) * d
p = np.array(np.meshgrid(xx, yy, 0, indexing="ij"))
p[0, :, ::2] += (xx[1] - xx[0]) * 0.5

p = p.reshape(3, -1).T

pp = plane2mesh(p, mesh, u, v)

plot_data_on_vertices(mesh, u, ncolors=15)
mlab.points3d(*pp.T, scale_factor=0.01)



"""
Figure 2: Surface harmonics modes
==================================

First 15 surface harmonics of a triangle mesh. The index (and spatial frequency) increases from left to right, row by row. The hole and outer boundary have zero tangential derivative. The mesh discretization is shown in the magnified inset to the right.

"""
from bfieldtools.mesh_conductor import MeshConductor

import numpy as np

SAVE = False
SAVE_DIR = "./Surface harmonics/"


c = MeshConductor(
    mesh_file=SAVE_DIR + "arbitrary_surface.stl",
    process=True,
    basis_name="suh",
    N_suh=15,
    fix_normals=True,
)


T_x = 1.5 * np.pi / 2
T_z = -1.02 * np.pi
rotmat = np.array(
    [
        [np.cos(T_z), -np.sin(T_z), 0, 0],
        [np.sin(T_z), np.cos(T_z), 0, 0],
        [0, 0, 1, 0],
        [0, 0, 0, 1],
    ]
) @ np.array(
    [
        [1, 0, 0, 0],
        [0, np.cos(T_x), -np.sin(T_x), 0],
        [0, np.sin(T_x), np.cos(T_x), 0],
        [0, 0, 0, 1],
    ]
)


c.mesh.apply_transform(rotmat)


from mayavi import mlab
from mayavi.api import Engine

e = Engine()
e.start()


f = mlab.figure(None, bgcolor=(1, 1, 1), fgcolor=(0.5, 0.5, 0.5), size=(750, 900))

surfaces = c.suh_basis.plot(
    Nfuncs=c.basis.shape[1], Ncols=2, dist=0.1, colormap="RdBu", figure=f, ncolors=256
)

f.scene.z_plus_view()
1 + 1
#
for idx, surface in enumerate(surfaces):
    surface.actor.mapper.interpolate_scalars_before_mapping = True
    surface.module_manager.scalar_lut_manager.data_range = np.array(
        [
            -np.max(np.abs(c.suh_basis.basis[:, idx])),
            np.max(np.abs(c.suh_basis.basis[:, idx])),
        ]
    )


# f.scene.camera.parallel_projection=1
# mlab.view(0,160)
f.scene.camera.zoom(1.2)
# f.scene.camera.roll(270)

if SAVE:
    mlab.savefig(
        SAVE_DIR + "surface_harmonics.png",
        figure=f,
        magnification=4,
    )

    mlab.close()

f = mlab.figure(None, bgcolor=(1, 1, 1), fgcolor=(0.5, 0.5, 0.5), size=(750, 600))

c.plot_mesh(representation="wireframe", figure=f)
c.plot_mesh(opacity=0.2, figure=f)
#
# f.scene.camera.parallel_projection=1
# mlab.view(0,160)
f.scene.camera.zoom(1.5)
f.scene.z_plus_view()
# f.scene.camera.roll(270)

if SAVE:
    mlab.savefig(
        SAVE_DIR + "suhmesh.png",
        figure=f,
        magnification=6,
    )
    mlab.close()



"""
B-field on the symmetry axis of a disc
===========================================

Compare a analytic expression and the numerical computation of
the B-field of a symmetric current distribution 
"""

import numpy as np
import matplotlib.pyplot as plt
import trimesh
from mayavi import mlab
from bfieldtools.mesh_magnetics import magnetic_field_coupling
from bfieldtools.mesh_conductor import MeshConductor, StreamFunction
from bfieldtools.utils import load_example_mesh


def field_disc(z, a):
    """Bfield z-component of streamfunction psi=r**2 z-axis
    for a disk with radius "a" on the xy plane
    """

    coeff = 1e-7
    field = -2 * (a**2 + 2 * z**2) / np.sqrt((a**2 + z**2)) + 4 * abs(z)
    field *= coeff * 2 * np.pi

    return field


# Load disc and subdivide
disc = load_example_mesh("unit_disc")
for i in range(4):
    disc = disc.subdivide()
disc.vertices -= disc.vertices.mean(axis=0)

# Stream functions
s = disc.vertices[:, 0] ** 2 + disc.vertices[:, 1] ** 2

#%% Create evaluation points and plot them
Np = 50
z = np.linspace(-1, 10, Np) + 0.001
points = np.zeros((Np, 3))
points[:, 2] = z
mlab.triangular_mesh(*disc.vertices.T, disc.faces, scalars=s, colormap="viridis")
mlab.points3d(*points.T, scale_factor=0.1)

bfield_mesh = magnetic_field_coupling(disc, points, analytic=True) @ s

#%% Plot asymptotic error with different radii
Nr = 100
drs = np.linspace(-0.0035, -0.0034, Nr)
from matplotlib import cm

colors = cm.viridis(np.linspace(0, 1, Nr))
err = []
for c, dr in zip(colors, drs):
    err.append(
        (abs((bfield_mesh[:, 2] - field_disc(z, 1 + dr)) / field_disc(z, 1 + dr)))[-1]
    )
plt.plot(1 + drs, err)
plt.ylabel("error")
plt.xlabel("disc radius")
#%% Determine the "most fit" radius and compare
plt.figure()
# Solve zero-crossing
pp = np.polyfit(drs[:30], err[:30], deg=1)
dr0 = -pp[1] / pp[0]

plt.semilogy(z, -field_disc(z, 1 + dr0))
plt.semilogy(z, -bfield_mesh[:, 2])
plt.xlabel("$z$")
plt.ylabel("$B_z$")


plt.figure()
# Plot the relative error
plt.plot(z, (abs(bfield_mesh[:, 2] - field_disc(z, 1 + dr0)) / field_disc(z, 1 + dr0)))
plt.xlabel("$z$")
plt.ylabel("relative error")



"""
Contains functions for computing vector calculus quantities on triangle surface meshes.
These include the gradient, rotated gradient, divergence, curl, Laplacian and mass matrices.
"""

__all__ = [
    "curl",
    "curl_matrix",
    "divergence",
    "divergence_matrix",
    "gradient",
    "gradient_matrix",
    "laplacian_matrix",
    "mass_matrix",
]

import numpy as np
from scipy.sparse import csr_matrix, spdiags


def laplacian_matrix(mesh, material_param=None):
    """
    Sparse Laplace-Beltrami operator.

    This function returns the so-called cotangent-Laplacian L, which can be
    used to solve the Laplacian eigenvectors as L @ v_i = -k_i^2 M @ v_i,
    where M is the mass matrix of the mesh (see mass_matrix).

    For meshes with boundaries, the matrix corresponds to zero-Neumann
    boundary conditions (the natural boundary condtion).

    When solving something with zero-Dirichlet condition
    only the inner vertices should be free parameters and
    the associated elements of the matrix should be used.

    For more discussion about the boundary conditions related
    to stream functions of divergence-free surface currents,
    see the associated publication about the physics (Makinen 2020).

    Parameters
    ----------
    mesh: Trimesh Mesh object
    material_param: array-like with length N_triangles
        material parameter for each triangle, used for resistance matrix

    Returns
    -------
    sparse csr_matrix (N_vertices, N_vertices)
        non-diagonal elements L_ij = - 0.5* (cot(alpha) + cot(beta))
        diagonal elements L_ii = - sum_{i!=j} L_ij

    """
    if material_param is None:
        p = 1
    else:
        p = material_param
    N = mesh.vertices.shape[0]
    R = mesh.vertices[mesh.faces]  # Nt x 3 (corners) x 3 (xyz)

    # Edges opposite to the vertex
    edges = np.roll(R, 1, -2) - np.roll(R, 2, -2)  # Nt x 3 (edges) x 3 (x,y,z)

    ii = []
    jj = []
    cot = []
    # Loop over edges in triangles
    for i in range(3):
        i1 = (i + 1) % 3
        i2 = (i + 2) % 3
        ii.append(mesh.faces[:, i1])
        jj.append(mesh.faces[:, i2])
        c = (
            -0.5
            * (edges[:, i1, :] * edges[:, i2, :]).sum(axis=-1)
            / (2 * mesh.area_faces)
        )
        # Append cot with cotangent terms multiplied by the material parameter
        cot.append(c * p)

    ii = np.ravel(ii)
    jj = np.ravel(jj)
    cot = np.ravel(cot)
    # Build the sparse matrix
    L = csr_matrix((cot, (ii, jj)), shape=(N, N), dtype=float)
    # Sum contribution from both triangles (alpha and beta angles)
    # neighbouring the edge
    L = L + L.T
    # Add the diagonal
    L = L - spdiags(L.sum(axis=0), 0, N, N)

    return L


def mass_matrix(mesh, lumped=False):
    """
    Computes mass matrix of a mesh.

    Parameters
    ----------
    mesh: Trimesh Mesh object
    lumped: Boolean
        If True, use lumped approximation of mass matrix. If False (default),
        compute exact matrix. See Reuter et al 2009, page 3 (DOI: 10.1016/j.cag.2009.03.005)

    Returns
    -------
    sparse csr_matrix of variable shape (see the description of laplacian_matrix)
        Mass matrix

    """

    if lumped:
        from .utils import dual_areas

        da = dual_areas(mesh.faces, mesh.area_faces)
        M = spdiags(da, 0, mesh.vertices.shape[0], mesh.vertices.shape[0]).tocsr()
    else:
        N = mesh.vertices.shape[0]

        ii = []
        jj = []
        area = []
        # Loop over edges in triangles
        for i in range(3):
            i1 = (i + 1) % 3
            i2 = (i + 2) % 3
            ii.append(mesh.faces[:, i1])
            jj.append(mesh.faces[:, i2])
            area.append(mesh.area_faces / 12)

        ii = np.ravel(ii)
        jj = np.ravel(jj)
        area = np.ravel(area)
        # Build sparse matrix
        M = csr_matrix((area, (ii, jj)), shape=(N, N), dtype=float)
        # Sum contribution from both triangles (alpha and beta angles)
        # neighbouring the edge
        M = M + M.T
        M = M + spdiags(M.sum(axis=0), 0, N, N)

    return M


def gradient_matrix(mesh, rotated=False):
    """
    Calculate a (rotated) gradient matrix for hat basis functions
    (stream functions) in a mesh.

    Parameters
    ----------
    mesh: Trimesh mesh object
    rotated: boolean
        If True, rotate gradient 90 degrees clockwise

    Returns
    -------
    3 sparse csr_matrices
    Gx ,Gy, Gx (Ntris, Nverts) matrices
        for calculating the components of gradient at triangles
    """
    R = mesh.vertices[mesh.faces]  # Nt x 3 (corners) x 3 (xyz)
    # Calculate edge vectors for each triangle
    edges = np.roll(R, 1, -2) - np.roll(R, 2, -2)  # Nt x 3 (edges) x 3 (x,y,z)

    tri_data = edges / (2 * mesh.area_faces[:, None, None])
    if not rotated:
        # Rotate 90 degrees CW to get the original gradient
        tri_data = np.cross(mesh.face_normals[:, None, :], tri_data, axis=-1)

    tri_data = tri_data.reshape(-1, 3).T
    ii = np.array(
        [[i] * 3 for i in range(len(mesh.faces))]
    ).ravel()  # [0,0,0,1,1,1,...]
    jj = mesh.faces.ravel()  # [t[0,0], t[0,1], t[0,2], t[1,0], t[1,1], t[1,2], ...]
    Gx = csr_matrix(
        (tri_data[0], (ii, jj)),
        shape=(mesh.faces.shape[0], mesh.vertices.shape[0]),
        dtype=float,
    )
    Gy = csr_matrix(
        (tri_data[1], (ii, jj)),
        shape=(mesh.faces.shape[0], mesh.vertices.shape[0]),
        dtype=float,
    )
    Gz = csr_matrix(
        (tri_data[2], (ii, jj)),
        shape=(mesh.faces.shape[0], mesh.vertices.shape[0]),
        dtype=float,
    )
    return Gx, Gy, Gz


def gradient(vals, mesh, rotated=False):
    """
    Applies mesh (rotated) gradient matrix operator on vector that is
    defined in the vertex locations of the mesh.

    Parameters
    ----------
    vals: Nv x 1 array of scalar data to compute the gradient of
    mesh: Trimesh object describing the triangular mesh
    rotated: boolean
        If True, rotate gradient 90 degrees clockwise

    Returns
    -------
    ndarray (3, Ntris)
        surface gradient of vals on each triangle

    """
    Gx, Gy, Gz = gradient_matrix(mesh, rotated)
    return np.array([Gx @ vals, Gy @ vals, Gz @ vals])


def divergence_matrix(mesh):
    """
    Divergence of tangential vector field on mesh faces as a linear mapping.

    Parameters
    ----------
    mesh: Trimesh object

    Returns
    -------
    3 sparse csc_matrices
        Dx ,Dy, Dz (Nverts, Ntri) matrices

    """
    from .utils import tri_normals_and_areas

    Gx, Gy, Gz = gradient_matrix(mesh, rotated=False)
    n, a = tri_normals_and_areas(mesh.vertices, mesh.faces)
    A = spdiags(a, 0, a.shape[0], a.shape[0])
    return -Gx.T * A, -Gy.T * A, -Gz.T * A


def curl_matrix(mesh):
    """
    Adjoint curl of tangential vector field.

    Parameters
    ----------
    mesh: Trimesh object

    Returns
    -------
    3 sparse csc_matrices
        Cx ,Cy, Cz (Nverts, Ntri) matrices

    """
    from .utils import tri_normals_and_areas

    Gx, Gy, Gz = gradient_matrix(mesh, rotated=True)
    n, a = tri_normals_and_areas(mesh.vertices, mesh.faces)
    A = spdiags(a, 0, a.shape[0], a.shape[0])
    return -Gx.T * A, -Gy.T * A, -Gz.T * A


def divergence(vecs, mesh):
    """
    Divergence mapping applied to tangential vector field 'vecs'

    Parameters
    ----------

    vecs: ndarray (3, Nfaces)
            vector field at mesh faces

    Returns
    -------
    ndarray (Nverts,)
        Divergence applied on the vector field

    """
    Dx, Dy, Dz = divergence_matrix(mesh)
    return Dx @ vecs[:, 0] + Dy @ vecs[:, 1] + Dz @ vecs[:, 2]


def curl(vecs, mesh):
    """
    Curl applied to tangential vector field

    Parameters
    ----------
    mesh: Trimesh object

    Returns
    -------
    ndarray (Nverts,)
        Curl applied on the vector field


    """
    Cx, Cy, Cz = curl_matrix(mesh)
    return Cx @ vecs[:, 0] + Cy @ vecs[:, 1] + Cz @ vecs[:, 2]



"""
Field distortion by a perfect mu-metal shield
===========================================================
"""

import numpy as np
import trimesh
from mayavi import mlab

from bfieldtools.mesh_conductor import MeshConductor, StreamFunction
from bfieldtools.mesh_calculus import gradient
from bfieldtools.utils import load_example_mesh, combine_meshes

# This doesn't matter, the problem is scale-invariant
scaling_factor = 1


planemesh = load_example_mesh("10x10_plane_hires")
planemesh.apply_scale(scaling_factor)

# Specify coil plane geometry
center_offset = np.array([9, 0, 0]) * scaling_factor
standoff = np.array([0, 4, 0]) * scaling_factor

# Create coil plane pairs
coil_plus = trimesh.Trimesh(
    planemesh.vertices + center_offset + standoff, planemesh.faces, process=False
)

coil_minus = trimesh.Trimesh(
    planemesh.vertices + center_offset - standoff, planemesh.faces, process=False
)

joined_planes = combine_meshes((coil_plus, coil_minus))
planemesh = joined_planes

# Create mesh class object
coil = MeshConductor(mesh_obj=joined_planes, fix_normals=True)

# Separate object for shield geometry
shieldmesh = load_example_mesh("closed_cylinder_remeshed")
shieldmesh.apply_scale(15)

shield = MeshConductor(mesh_obj=shieldmesh, process=True, fix_normals=True)

#%%
N = 80
points = np.zeros((N * N, 3))
w = 12
X, Y = np.meshgrid(np.linspace(-w, w, N), np.linspace(-w, w, N), indexing="ij")
X += planemesh.vertices.mean(axis=0)[0]
Y += planemesh.vertices.mean(axis=0)[1]
points[:, 0] = X.flatten()
points[:, 1] = Y.flatten()
points[:, 2] += planemesh.vertices.mean(axis=0)[2]


def plot_plane():
    mlab.triangular_mesh(
        [X.min(), X.max(), X.max(), X.min()],
        [Y.min(), Y.min(), Y.max(), Y.max()],
        np.zeros(4),
        np.array([[0, 1, 2], [2, 3, 0]]),
        color=(0.1, 0.1, 0.1),
        opacity=0.6,
    )


#%%
fig = mlab.figure(bgcolor=(1, 1, 1))
s0 = mlab.triangular_mesh(
    *shieldmesh.vertices.T, shieldmesh.faces, color=(0.5, 0.5, 0.5), opacity=0.3
)
s0.actor.property.backface_culling = False
s0.actor.property.ambient = 0.5

I_prim = np.load("../publication_software/Shielded coil/biplanar_streamfunction.npy")
sprim = StreamFunction(coil.vert2inner @ I_prim, coil)
m = max(abs(sprim))
s1 = sprim.plot(False, 16, vmin=-m, vmax=m)
s2 = sprim.plot(True, 20)
s2.actor.mapper.scalar_visibility = False
s2.actor.property.line_width = 1.2
# s2 = mlab.triangular_mesh(*planemesh.vertices.T, planemesh.faces, scalars=I_prim,
#                          colormap='RdBu')
# s2.enable_contours = True
# s2.contour.filled_contours = True
# s2.contour.number_of_contours = 20
s2.actor.property.render_lines_as_tubes = True
plot_plane()
scene = s0.module_manager
scene.scene.camera.position = [
    39.154871143623325,
    -40.509425675368334,
    26.56155776567048,
]
scene.scene.camera.focal_point = [
    4.239945673839333,
    1.041549923485209,
    -0.0005302515738243585,
]
scene.scene.camera.view_angle = 30.0
scene.scene.camera.view_up = [
    -0.28276020498745635,
    0.33658483701858727,
    0.898196701154387,
]
scene.scene.camera.clipping_range = [16.139073445910277, 116.31572537292347]
scene.scene.camera.compute_view_plane_normal()
scene.scene.render()


#%%
#%% Calculate primary potential matrix
# Compute slightly inside
d = 1e-3
P_prim = coil.U_coupling(shieldmesh.vertices - d * shieldmesh.vertex_normals)


#%% Plot the resulting primary potential
mlab.figure()
s = mlab.triangular_mesh(
    *shieldmesh.vertices.T, shieldmesh.faces, scalars=P_prim @ sprim, opacity=1.0
)
s.enable_contours = True
s.contour.filled_contours = True
s.contour.number_of_contours = 30

#%%
#%% Calculate linear collocation BEM matrix
P_shield = shield.U_coupling(shieldmesh.vertices - d * shieldmesh.vertex_normals)

#%%
#%% Solve equivalent stream function for the perfect linear mu-metal layer
I_shield = np.linalg.solve(-P_shield, P_prim @ sprim)
# I_shield = P_prim @ I_prim
s_shield = StreamFunction(I_shield, shield)
g = gradient(s_shield, shieldmesh, rotated=True)

#%% Plot the result
fig = mlab.figure(bgcolor=(1, 1, 1))
s0 = mlab.triangular_mesh(
    *shieldmesh.vertices.T, shieldmesh.faces, color=(0.5, 0.5, 0.5), opacity=0.3
)
s0.actor.property.backface_culling = False
s1 = s_shield.plot(False, 256)
# s1.actor.property.opacity=0.8
s1.actor.property.backface_culling = False
# s2 = s_shield.plot(True, 10)
mlab.quiver3d(
    *shieldmesh.triangles_center.T,
    *g,
    color=(1, 1, 1),
    mode="arrow",
    scale_factor=0.0000008,
    scale_mode="vector"
)
# s1.contour.filled_contours = True
# s1.contour.number_of_contours = 30
# s2.actor.property.render_lines_as_tubes = True
# s1.actor.property.ambient = 0.2

scene = s1.module_manager
plot_plane()
scene.scene.camera.position = [
    39.154871143623325,
    -40.509425675368334,
    26.56155776567048,
]
scene.scene.camera.focal_point = [
    4.239945673839333,
    1.041549923485209,
    -0.0005302515738243585,
]
scene.scene.camera.view_angle = 30.0
scene.scene.camera.view_up = [
    -0.28276020498745635,
    0.33658483701858727,
    0.898196701154387,
]
scene.scene.camera.clipping_range = [16.139073445910277, 116.31572537292347]
scene.scene.camera.compute_view_plane_normal()
scene.scene.render()


#%%
U2_shield = shield.U_coupling(points)
U2_prim = coil.U_coupling(points)


#%%
from bfieldtools.contour import scalar_contour

cc1 = scalar_contour(planemesh, planemesh.vertices[:, 2], contours=[-0.1])
cc1 = np.vstack(cc1)
cc1a = cc1[: cc1.shape[0] // 2]
cc1b = cc1[cc1.shape[0] // 2 :]
cc2 = scalar_contour(shieldmesh, shieldmesh.vertices[:, 2], contours=[-0.1])
cc2 = np.vstack(cc2)
cc2a = cc1[: cc2.shape[0] // 2]
cc2b = cc1[cc2.shape[0] // 2 :]

import matplotlib.pyplot as plt
import matplotlib.colors as colors


def truncate_colormap(cmap, minval=0.0, maxval=1.0, n=256):
    new_cmap = colors.LinearSegmentedColormap.from_list(
        "trunc({n},{a:.2f},{b:.2f})".format(n=cmap.name, a=minval, b=maxval),
        cmap(np.linspace(minval, maxval, n)),
    )
    return new_cmap


cmap = plt.get_cmap("RdBu")
# cmap.set_over((0.95,0.95,0.95))
# cmap.set_under((0.95,0.95,0.95))

u0 = abs(
    np.sum(U2_shield, axis=1).reshape(N, N)
)  # Solid angle of the shield, zero outside
u0 /= u0.max()
u0[u0 < 1e-6] = 0
u1 = (U2_prim @ sprim).reshape(N, N)
u2 = (U2_shield @ I_shield).reshape(N, N) * u0
u3 = (u1 + u2) * u0

vmax = np.max(abs(u3)) * 0.99
levels = np.linspace(-vmax, vmax, 120)
levels = np.hstack((-np.max(abs(u3)), levels, np.max(abs(u3))))
p = plt.contourf(
    X,
    Y,
    u1,
    levels=levels,
    cmap=cmap,
    vmin=-vmax,
    vmax=vmax,
    norm=colors.SymLogNorm(linthresh=0.2 * vmax, linscale=0.8, vmin=-vmax, vmax=vmax),
)
plt.plot(cc1a[:, 0], cc1a[:, 1], linewidth=3, color="gray")
plt.plot(cc1b[:, 0], cc1b[:, 1], linewidth=3, color="gray")
plt.axis("image")
plt.axis("off")
xlims = p.ax.get_xlim()
plt.figure()
p = plt.contourf(
    X,
    Y,
    u2,
    levels=levels,
    cmap=cmap,
    vmin=-vmax,
    vmax=vmax,
    norm=colors.SymLogNorm(linthresh=0.2 * vmax, linscale=0.8, vmin=-vmax, vmax=vmax),
)
plt.plot(cc1a[:, 0], cc1a[:, 1], linewidth=3, color="gray")
plt.plot(cc1b[:, 0], cc1b[:, 1], linewidth=3, color="gray")
plt.plot(cc2[:, 0], cc2[:, 1], linewidth=3, color="gray")
plt.axis("image")
plt.axis("off")
p.ax.set_xlim(xlims)
plt.figure()
p = plt.contourf(
    X,
    Y,
    u3,
    levels=levels,
    cmap=cmap,
    vmin=-vmax,
    vmax=vmax,
    norm=colors.SymLogNorm(linthresh=0.2 * vmax, linscale=0.8, vmin=-vmax, vmax=vmax),
)
plt.plot(cc1a[:, 0], cc1a[:, 1], linewidth=3, color="gray")
plt.plot(cc1b[:, 0], cc1b[:, 1], linewidth=3, color="gray")
plt.plot(cc2[:, 0], cc2[:, 1], linewidth=3, color="gray")
plt.axis("image")
plt.axis("off")
p.ax.set_xlim(xlims)



from bfieldtools import quadratures
from bfieldtools import utils

from .test_line_conductor import _fake_line_conductor


import pytest
import numpy as np


@pytest.mark.parametrize(
    "rule", quadratures.dunavant_rules + quadratures.gauss_legendre_rules
)
def test_rules(rule):
    assert np.allclose(rule.weights.sum(), 1)
    assert np.allclose(rule.points.sum(axis=1), 1)


@pytest.mark.parametrize(
    "method", [("dunavant", 0), ("dunavant", 1), ("dunavant", 2), ("dunavant", 3)]
)
def test_test_integration(method):
    val = 5
    verts = np.array([[0, 0, 0], [1, 0, 0], [0, 1, 0]])
    tris = np.array([[0, 1, 2]])

    def uniform(r):
        return val * np.ones(r.shape[:2])

    w, qp = quadratures.get_quad_points(verts, tris, method)
    assert np.allclose(uniform(qp) @ w, val)


def test_quad_points():
    mesh = utils.load_example_mesh("unit_disc")
    w, qp = quadratures.get_quad_points(mesh.vertices, mesh.faces, ("dunavant", 2))


def test_dunavant3():
    p = np.array(
        [
            [1 - 2 * 0.445948490915965, 0.445948490915965, 0.445948490915965],
            [0.445948490915965, 1 - 2 * 0.445948490915965, 0.445948490915965],
            [0.445948490915965, 0.445948490915965, 1 - 2 * 0.445948490915965],
            [1 - 2 * 0.091576213509771, 0.091576213509771, 0.091576213509771],
            [0.091576213509771, 1 - 2 * 0.091576213509771, 0.091576213509771],
            [0.091576213509771, 0.091576213509771, 1 - 2 * 0.091576213509771],
        ]
    )
    pq = quadratures.get_triangle_rule(("dunavant", 3)).points
    assert np.allclose(pq, p)


def test_line_quad_points():
    lp = _fake_line_conductor()

    w, qp = quadratures.get_line_quad_points(lp.vertices[lp.entities[0].points])
    w0, qp0 = quadratures.get_line_quad_points(
        lp.vertices[lp.entities[0].points], "midpoint"
    )
    w1, qp1 = quadratures.get_line_quad_points(
        lp.vertices[lp.entities[0].points], ("gauss_legendre", 0)
    )

    assert np.allclose(w0, w1)
    assert np.allclose(qp0, qp1)

    w, qp = quadratures.get_line_quad_points(
        lp.vertices[lp.entities[0].points], "trapezoid"
    )
    w, qp = quadratures.get_line_quad_points(
        lp.vertices[lp.entities[0].points], "simpson"
    )
    w, qp = quadratures.get_line_quad_points(
        lp.vertices[lp.entities[0].points], ("gauss_legendre", 2)
    )


def test_spherical_quad_points():
    qp = quadratures.get_spherical_quad_points()
    assert np.allclose(sum(qp.weights), 1)
    for coord in qp.points.T:
        assert np.allclose(coord.sum(), 0, atol=1e-7)



import numpy as np

from bfieldtools.line_magnetics import magnetic_field
from bfieldtools.line_magnetics import scalar_potential
from bfieldtools.line_magnetics import vector_potential


def create_3d_grid(xx, yy, zz):
    """Creates a direct product grid from three 1D arrays (xx, yy and zz)
    that is appropriately formated for `scalar_potential` and `magnetic_field`.
    """
    X, Y, Z = np.meshgrid(xx, yy, zz, indexing="ij")

    x = X.ravel()
    y = Y.ravel()
    z = Z.ravel()

    return np.array([x, y, z]).T


def test_field_calculations_in_grid():
    """
    The test checks the relation B = -mu0*grad(U) between the scalar potential and
    the magnetic field, and B = rot(A) between the vector potential and the magnetic
    field produced by a rectangular current loop. It compares magnetic fields
    computed by three different methods and prints out a relative error and a
    warning if the error is too large.

    magnetic_field, scalar_potential and vector_potential use independent analytic
    formulas and it is useful to check their consistency.

    The test also compares the magnetic field of a rectangular loop to values
    computed externally and prints a notification if the discrepancy is found
    to be too large.
    """
    # Defines a rectangular current loop offset by ly from the xz plane
    length = 0.3  # (m)
    width = 6e-2  # (m)
    ly = 0  # (m)
    loop_points = np.array(
        [
            [width / 2, ly, length / 2],
            [width / 2, ly, -length / 2],
            [-width / 2, ly, -length / 2],
            [-width / 2, ly, length / 2],
            [width / 2, ly, length / 2],
        ]
    )

    # Defines a grid inside a 3D box
    nx = ny = nz = 10

    # Box sizes
    lx = ly = lz = 0.4  # (m)

    xx = np.linspace(-lx / 2, lx / 2, nx)
    yy = np.linspace(-ly / 2, ly / 2, ny)
    zz = np.linspace(-lz / 2, lz / 2, nz)

    grid = create_3d_grid(xx, yy, zz)

    d = 1e-6  # Numerical differentiation step (m)

    # Defines offseted grids for numerical differentiation
    grid_xp = create_3d_grid(xx + d, yy, zz)
    grid_yp = create_3d_grid(xx, yy + d, zz)
    grid_zp = create_3d_grid(xx, yy, zz + d)

    grid_xm = create_3d_grid(xx - d, yy, zz)
    grid_ym = create_3d_grid(xx, yy - d, zz)
    grid_zm = create_3d_grid(xx, yy, zz - d)

    # The scalar potential at r+-dx, r+-dy and r+-dz
    U_xp = scalar_potential(loop_points, grid_xp)
    U_xm = scalar_potential(loop_points, grid_xm)
    U_yp = scalar_potential(loop_points, grid_yp)
    U_ym = scalar_potential(loop_points, grid_ym)
    U_zp = scalar_potential(loop_points, grid_zp)
    U_zm = scalar_potential(loop_points, grid_zm)

    # The vector potential at r+-dx, r+-dy and r+-dz
    A_xp = vector_potential(loop_points, grid_xp)
    A_xm = vector_potential(loop_points, grid_xm)
    A_yp = vector_potential(loop_points, grid_yp)
    A_ym = vector_potential(loop_points, grid_ym)
    A_zp = vector_potential(loop_points, grid_zp)
    A_zm = vector_potential(loop_points, grid_zm)

    # The calculation of magnetic field directly and via scalar potential and
    # compare the two methods

    mu0 = 4 * np.pi * 1e-7

    # Compute the magnetic field from scalar potential, B = -mu0*U
    B = -(mu0 / (2 * d)) * np.vstack([(U_xp - U_xm), (U_yp - U_ym), (U_zp - U_zm)]).T

    # Compute the madnetic field directly
    B_2 = magnetic_field(loop_points, grid)

    # Compute the magnetic field from vector potential, B = rot(A)
    B_3 = (
        1
        / (2 * d)
        * np.vstack(
            [
                (A_yp[:, 2] - A_ym[:, 2]) - (A_zp[:, 1] - A_zm[:, 1]),
                -(A_xp[:, 2] - A_xm[:, 2]) + (A_zp[:, 0] - A_zm[:, 0]),
                (A_xp[:, 1] - A_xm[:, 1]) - (A_yp[:, 0] - A_ym[:, 0]),
            ]
        ).T
    )

    error_bounds = [5e-8, 2e-7, 1e-9]

    # Print the maximum relative deviation between the fields calculated by
    # first two methods.
    err0 = np.max(np.abs((B_2 - B) / (np.abs(B) + np.abs(B_2))))

    if err0 < error_bounds[0]:
        print("The scalar potential test is passed.")
    else:
        print(
            "The scalar potential test is failed. The relative error is expected "
            "to be around 3.6e-8."
        )
    print(f"Relative error: {err0}\n")

    err1 = np.max(np.abs((B_2 - B_3) / (np.abs(B_3) + np.abs(B_2))))

    if err1 < error_bounds[1]:
        print("The vector potential test is passed.")
    else:
        print(
            "The vector potential test is failed. The relative error is expected "
            "to be around 1e-7."
        )
    print(f"Relative error: {err1}\n")

    # Calsulates the magnetic field on the diagonal of the grid.
    B_lin = magnetic_field(loop_points, np.vstack([xx, yy, zz]).T)

    # B_ref is magnetic field of the same configuration of currents computed using
    # a different external program.
    B_ref = [
        [5.0405815897640994e-8, 5.692103512792891e-9, 3.786395675311007e-8],
        [1.130657032079328e-7, 1.776261166601295e-8, 7.131528630113071e-8],
        [3.2247298177571e-7, 6.873374013470318e-8, 1.3724305960053128e-7],
        [1.2321797579182563e-6, 3.0821577841363554e-7, 1.7259933917883818e-7],
        [6.628725320025185e-6, 6.310902639454654e-6, 3.273181280102712e-8],
        [6.628725320025185e-6, 6.310902639454654e-6, 3.273181280102712e-8],
        [1.2321797579182563e-6, 3.0821577841363554e-7, 1.7259933917883818e-7],
        [3.2247298177571e-7, 6.873374013470318e-8, 1.3724305960053128e-7],
        [1.130657032079328e-7, 1.776261166601295e-8, 7.131528630113071e-8],
        [5.0405815897640994e-8, 5.692103512792891e-9, 3.786395675311007e-8],
    ]

    err2 = np.max(np.abs((B_lin - B_ref) / (np.abs(B_lin) + np.abs(B_ref))))

    if err2 < error_bounds[2]:
        print("The magnetic field test is passed.")
    else:
        print(
            "The magnetic field test is failed. The relative error is expected "
            "to be below 1e-9."
        )
    print(f"Relative error: {err2}")

    assert err0 < error_bounds[0] and err1 < error_bounds[1] and err2 < error_bounds[2]



"""
Functions for constructing real spherical harmonics (Ylms), their gradients
and related magnetic field 'basis vectorfunctions'
(Wlms for r**l components, Vlms for r**(-l) components).

Uses notations and definitions by Plattner and Simons (2014; https://arxiv.org/pdf/1306.3201.pdf)
and the same normalization conventions.

Integration over a surface of unit sphere is
used as the inner product <C,D> = int C dot D dOmega.

Also has many of functions for spherical <-> cartesian transformations.
"""

__all__ = [
    "Blm",
    "Plm",
    "Rotmatrix",
    "SphBasis",
    "Vlm",
    "Wlm",
    "Xlm",
    "basis_fields",
    "basis_potentials",
    "cartesian2spherical",
    "cartvec2sph",
    "compute_sphcoeffs_mesh",
    "derlpmn_em",
    "derxlm",
    "dphiylm",
    "dthylm",
    "field",
    "lpmn_em",
    "plotBVlm_volume",
    "plotBWlm_volume",
    "plotVlm",
    "plotWlm",
    "plotXlm",
    "plotYlm",
    "plotYlms",
    "potential",
    "sinxlm",
    "spherical2cartesian",
    "sphvec2cart",
    "xlm",
    "ylm",
]

import numpy as np
from scipy.special import factorial, lpmn

from .quadratures import get_spherical_quad_points

from .mesh_calculus import gradient_matrix
from .utils import tri_normals_and_areas


############################################
# Helper functions for sph computation


def cartesian2spherical(p, zaxis_approx=True):
    """
    Maps cartesian coordinates to spherical.

    Parameters
    ----------
    p: Nx3 array
        cartesian coordinates
    zaxis_approx: Boolean (True)
        If True, apply regularization that avoids singularity on z-axis

    Returns
    -------

    sp: Nx3 array
        spherical coordinates [r, theta, phi]

    """

    r = np.sqrt(p[:, 0] ** 2 + p[:, 1] ** 2 + p[:, 2] ** 2)
    theta = np.arctan2(np.sqrt(p[:, 0] ** 2 + p[:, 1] ** 2), p[:, 2])
    phi = np.arctan2(p[:, 1], p[:, 0])

    sp = np.array((r, theta, phi)).T

    if zaxis_approx:
        eps = 1e-6
        theta_mask_0 = np.abs(sp[:, 1]) < eps
        theta_mask_pi = np.abs(sp[:, 1] - np.pi) < eps
        sp[theta_mask_0, 1] += eps
        sp[theta_mask_pi, 1] -= eps

    return sp


def spherical2cartesian(sp):
    """
    Maps spherical coordinates to cartesian.

    Parameters
    ----------
    sp: Nx3 array
        spherical coordinates [r, theta, phi]

    Returns
    -------
    p: Nx3 array
        cartesian croodinates
    """

    X = sp[:, 0] * np.sin(sp[:, 1]) * np.cos(sp[:, 2])
    Y = sp[:, 0] * np.sin(sp[:, 1]) * np.sin(sp[:, 2])
    Z = sp[:, 0] * np.cos(sp[:, 1])

    p = np.array((X, Y, Z)).T
    return p


def Rotmatrix(sp):
    """
    Constructs rotation matrix from cartesian coordinates to spherical.

    Parameters
    ----------
    sp: Nx3 array
        spherical coordinates [r, theta, phi]

    Returns
    -------
    vmat: 3x3 array
        rotation matrix from cartesian to spherical.
    """

    vmat = np.zeros((3, 3))
    vmat[0, 0] = np.sin(sp[1]) * np.cos(sp[2])
    vmat[0, 1] = np.sin(sp[1]) * np.sin(sp[2])
    vmat[0, 2] = np.cos(sp[1])
    vmat[1, 0] = np.cos(sp[1]) * np.cos(sp[2])
    vmat[1, 1] = np.cos(sp[1]) * np.sin(sp[2])
    vmat[1, 2] = -1 * np.sin(sp[1])
    vmat[2, 0] = -1 * np.sin(sp[2])
    vmat[2, 1] = np.cos(sp[2])
    vmat[2, 2] = 0

    return vmat


def cartvec2sph(sp, vec):
    """
    Transforms cartesian vector to spherical coordinates.

    Parameters
    ----------
    sp: Nx3 array
        spherical coordinates [r, theta, phi]
    vec: Nx3 array
        vector in cartesian coordinates [e_x, e_y, e_z]

    Returns
    -------
    svec: Nx3 array
        vector in spherical coordinates [e_r, e_theta, e_phi]

    """

    svec = np.zeros(vec.shape)
    for i in range(sp.shape[0]):
        vmat = Rotmatrix(sp[i])
        svec[i] = vmat @ vec[i]
    return svec


def sphvec2cart(sp, vec):
    """
    Transforms cartesian vector to spherical coordinates.

    Parameters
    ----------
    sp: Nx3 array
        spherical coordinates [r, theta, phi]
    vec: Nx3 array
        vector in spherical coordinates [e_r, e_theta, e_phi]

    Returns
    -------
    svec: Nx3 array
        vector in cartesian coordinates [e_x, e_y, e_z]

    """

    svec = np.zeros(vec.shape)
    for i in range(sp.shape[0]):
        vmat = Rotmatrix(sp[i])
        svec[i] = np.transpose(vmat) @ vec[i]

    return svec


############################################
# Functions for generating Legendre polynomials,
# real spherical harmonics and vector spherical harmonics


def lpmn_em(l, m, x):
    """
    Computes associated Legendre function (Plm) of the first kind of order m and degree l.

    Parameters
    ----------

    l: int
        degree of Plm
    m: int
        order of Plm
    x: Nx1 array
        evaluation points

    Returns
    -------
    lp: Nx1 array
        Plm at `x`

    """

    lp = np.zeros(x.shape)
    for i in range(x.shape[0]):
        a, b = lpmn(m, l, x[i])
        lp[i] = a[np.abs(m), l]
    return lp


def derlpmn_em(l, m, x):
    """
    Computes derivative of associated Legendre function (Plm) of the first kind of order m and degree l
    with respect to the argument x.

    Parameters
    ----------
    l: int
        degree of Plm
    m: int
        order of Plm
    x: Nx1 array
        evaluation points

    Returns
    -------
    derlp: Nx1 array
        dPlm/dx at `x`

    """

    derlp = np.zeros(x.shape)
    for i in range(x.shape[0]):
        a, b = lpmn(m, l, x[i])
        derlp[i] = b[np.abs(m), l]
    return derlp


def xlm(l, m, theta):
    """
    Xlm-function used in the definition of spherical harmonics (Ylm).
    Follows notation of Plattner and Simons (2014);
    see Eqs. 1--3 in https://arxiv.org/pdf/1306.3201.pdf.

    Parameters
    ----------
    l: int
        degree of Xlm
    m: int
        order of Xlm
    theta: Nx1 array
        evaluation points, theta at spherical coordinates

    Returns
    -------
    Xlm: Nx1 array
        Xlm at theta

    """

    xlm = ((2 * l + 1) / (4 * np.pi)) ** 0.5 * (
        factorial(l - m) / factorial(l + m)
    ) ** 0.5
    xlm *= lpmn_em(l, m, np.cos(theta))
    return xlm


def ylm(l, m, theta, phi):
    """
    Real spherical harmonics as defined by Plattner and Simons (2014);
    see Eqs. 1--3 in https://arxiv.org/pdf/1306.3201.pdf.

    Parameters
    ----------
    l: int
        degree of Ylm
    m: int
        order of Ylm
    theta: Nx1 array
        evaluation points, theta at spherical coordinates
    phi: Nx1 array
        evaluation points, phi at spherical coordinates

    Returns
    -------
    Ylm: Nx1 array
        Ylm at (theta,phi)

    """

    if m < 0:
        ylm = np.sqrt(2) * xlm(l, m, theta) * np.cos(m * phi)
    elif m == 0:
        ylm = xlm(l, 0, theta)
    elif m > 0:
        ylm = np.sqrt(2) * xlm(l, m, theta) * np.sin(m * phi)

    return ylm


def derxlm(l, m, theta):
    """
    Derivative of Xlm with respect to theta.

    Parameters
    ----------
    l: int
        degree of Xlm
    m: int
        order of Xlm
    theta: Nx1 array
        evaluation points, theta at spherical coordinates

    Returns
    -------
    derxlm: Nx1 array
        dXlm/dtheta at theta

    """

    derxlm = ((2 * l + 1) / (4 * np.pi)) ** 0.5 * (
        factorial(l - m) / factorial(l + m)
    ) ** 0.5
    derxlm *= derlpmn_em(l, m, np.cos(theta))
    derxlm *= -1 * np.sin(
        theta
    )  # this comes from dXlm(cos(theta))/dtheta = dXlm(cos(theta))/dcos(theta)*(-sin(theta))
    return derxlm


def sinxlm(l, m, theta):
    """
    Computes m/(sin(theta))*Xlm.

    Parameters
    ----------
    l: int
        degree of Xlm
    m: int
        order of Xlm
    theta: Nx1 array
        evaluation points, theta at spherical coordinates

    Returns
    -------
    sinxlm: Nx1 array
        m/(sin(theta))*Xlm at theta

    """

    sinxlm = m / (np.sin(theta)) * xlm(l, m, theta)
    return sinxlm


def dthylm(l, m, theta, phi):
    """
    Derivative of Ylm with respect to theta dYlm/dtheta.

    Parameters
    ----------
    l: int
        degree of Ylm
    m: int
        order of Ylm
    theta: Nx1 array
        evaluation points, theta at spherical coordinates
    phi: Nx1 array
        evaluation points, phi at spherical coordinates

    Returns
    -------
    dthylm: Nx1 array
        dYlm/dtheta at (theta, phi).

    """

    if m < 0:
        dthylm = np.sqrt(2) * derxlm(l, m, theta) * np.cos(m * phi)
    elif m == 0:
        dthylm = derxlm(l, 0, theta)
    elif m > 0:
        dthylm = np.sqrt(2) * derxlm(l, m, theta) * np.sin(m * phi)
    return dthylm


def dphiylm(l, m, theta, phi):
    """
    Derivative of Ylm with respect to phi dYlm/dphi.

    Parameters
    ----------
    l: int
        degree of Ylm
    m: int
        order of Ylm
    theta: Nx1 array
        evaluation points, theta at spherical coordinates
    phi: Nx1 array
        evaluation points, phi at spherical coordinates

    Returns
    -------
    dphiylm: Nx1 array
        dYlm/dphi at (theta, phi).

    """

    if m < 0:
        dphiylm = -np.sqrt(2) * np.sin(m * phi) * sinxlm(l, m, theta)
    elif m == 0:
        dphiylm = 0
    elif m > 0:
        dphiylm = np.sqrt(2) * np.cos(m * phi) * sinxlm(l, m, theta)
    return dphiylm


def Plm(l, m, theta, phi):
    """
    Plm vector function (see Eq. 18 Plattner and Simons (2014)).

    Parameters
    ----------
    l: int
        degree of Plm
    m: int
        order of Plm
    theta: Nx1 array
        evaluation points, theta at spherical coordinates
    phi: Nx1 array
        evaluation points, phi at spherical coordinates

    Returns
    -------
    Plm: Nx3 array
        Plm at (theta, phi).

    """

    Plm = np.zeros((theta.shape[0], 3))
    Plm[:, 0] = ylm(l, m, theta, phi)
    return Plm


def Blm(l, m, theta, phi):
    """
    Blm vector function (see Eq. 19 Plattner and Simons (2014)).

    Parameters
    ----------
    l: int
        degree of Plm
    m: int
        order of Plm
    theta: Nx1 array
        evaluation points, theta at spherical coordinates
    phi: Nx1 array
        evaluation points, phi at spherical coordinates

    Returns
    -------
    Blm: Nx3 array
        Blm at (theta, phi).

    """

    Blm = np.zeros((theta.shape[0], 3))

    Blm[:, 1] = dthylm(l, m, theta, phi)

    Blm[:, 2] = dphiylm(l, m, theta, phi)

    Blm *= 1 / np.sqrt(l * (l + 1))
    return Blm


def Wlm(l, m, theta, phi):
    """
    Vector basis function (Wlm) for r**l component of the magnetic field.
    Normalization <Wlm,Wkn> = delta_lk,mn.

    Parameters
    ----------
    l: int
        degree of Wlm
    m: int
        order of Wlm
    theta: Nx1 array
        evaluation points, theta at spherical coordinates
    phi: Nx1 array
        evaluation points, phi at spherical coordinates

    Returns
    -------
    Wlm: Nx3 array
        Wlm at (theta, phi).

    """

    Wlm = l * Plm(l, m, theta, phi) + np.sqrt(l * (l + 1)) * Blm(l, m, theta, phi)
    Wlm *= 1 / np.sqrt(2 * l**2 + l)
    return Wlm


def Vlm(l, m, theta, phi):
    """
    Vector basis function (Vlm) for r**(-l) component of the magnetic field.
    Normalization <Vlm,Vkn> = delta_lk,mn.

    Parameters
    ----------
    l: int
        degree of Vlm
    m: int
        order of Vlm
    theta: Nx1 array
        evaluation points, theta at spherical coordinates
    phi: Nx1 array
        evaluation points, phi at spherical coordinates

    Returns
    -------
    Vlm: Nx3 array
        Vlm at (theta, phi).

    """

    Vlm = -1 * (l + 1) * Plm(l, m, theta, phi) + np.sqrt(l * (l + 1)) * Blm(
        l, m, theta, phi
    )
    Vlm *= 1 / np.sqrt((l + 1) * (2 * l + 1))
    return Vlm


def Xlm(l, m, theta, phi):
    """
    Vector spherical harmonics basis function (Xlm).
    Normalization <Xlm,Xkn> = delta_lk,mn.

    Parameters
    ----------
    l: int
        degree of Xlm
    m: int
        order of Xlm
    theta: Nx1 array
        evaluation points, theta at spherical coordinates
    phi: Nx1 array
        evaluation points, phi at spherical coordinates

    Returns
    -------
    Xlm: Nx3 array
        Xlm at (theta, phi).

    """

    temp = Blm(l, m, theta, phi)
    r = np.zeros(temp.shape)
    r[:, 0] = 1
    Xlm = -1 * np.cross(r, temp, axis=1)
    return Xlm


############################################
# Potential and field from the spherical harmonics


def basis_potentials(p, lmax, normalization="default", R=1):
    """
    Computes inner/outer basis functions for magnetic scalar potential.
    Ignores the 'DC' component l=0.

    Parameters
    ----------
    p: Nx3 array
        coordinates in which the potential is computed
    lmax: int
        maximum degree l which is used in computing
    normalization: string
        which normalization scheme to use
        "default": Integration over unit sphere <Y_lm,Y_l'm'> = delta_ll'mm' (default)
        "energy": each term in inner/outer basis function with respect to
        sphere with radius R is normalized to unit energy
    R: float
        Sphere radius that separates inner/outer components

    Returns
    -------
    pot1: Nxlmax*(lmax+2) array
        inner potential (alpha_lm) basis functions at p
    pot2: Nxlmax*(lmax+2) array
        outer potential (beta_lm) basis functions at p

    """
    mu0 = 1e-7 * 4 * np.pi
    L = lmax * (lmax + 2)

    pot1 = np.zeros((p.shape[0], L))
    pot2 = np.zeros((p.shape[0], L))

    sp = cartesian2spherical(p)

    lind = 0
    for l in range(1, lmax + 1):
        for m in range(-1 * l, l + 1):
            _ylm = ylm(l, m, sp[:, 1], sp[:, 2])

            pot1[:, lind] = sp[:, 0] ** (-l - 1) * _ylm
            pot2[:, lind] = sp[:, 0] ** l * _ylm

            if normalization == "energy":
                pot1[:, lind] *= np.sqrt(R ** (2 * l + 1) / ((l + 1) * mu0))
                pot2[:, lind] *= 1 / np.sqrt(R ** (2 * l + 1) * l * mu0)

            lind += 1

    return pot1, pot2


def potential(p, acoeffs, bcoeffs, lmax, normalization="default", R=1):
    """
    Computes magnetic scalar potential from the sph coefficients.
    Ignores the 'DC' component l=0.

    Parameters
    ----------
    p: Nx3 array
        coordinates in which the potential is computed
    acoeffs: lmax*(lmax+2)x1 array
        spectral coefficients of r**l terms
    bcoeffs: lmax*(lmax+2)x1 array
        spectral coefficients of r**(-l) terms
    lmax: int
        maximum degree l which is used in computing
    normalization: string
        which normalization scheme to use
        "default": Integration over unit sphere <Y_lm,Y_l'm'> = delta_ll'mm' (default)
        "energy": each term in inner/outer basis function with respect to
        sphere with radius R is normalized to unit energy
    R: float
        Sphere radius that separates inner/outer components

    Returns
    -------
    pot1: Nx array
        magnetic scalar potential at p

    """

    basis = basis_potentials(p, lmax, normalization, R)

    return basis[0] @ acoeffs + basis[1] @ bcoeffs


def field(p, acoeffs, bcoeffs, lmax, normalization="default", R=1):
    """
      Computes magnetic field at some point from the sph coefficients.
    Ignores the 'DC' component l=0.

    Parameters
    ----------
    p: Nx3 array
        coordinates in which the potential is computed
    acoeffs: lmax*(lmax+2)x1 array
        spectral coefficients of r**l terms
    bcoeffs: lmax*(lmax+2)x1 array
        spectral coefficients of r**(-l) terms
    lmax: int
        maximum degree l which is used in computing
    normalization: string
        "default": the fields correspond to normalized magnetic potential.
        "unit": the fields are normalized w.r.t integration over the unit sphere.
        "energy": each term in inner/outer basis function with respect to
        sphere with radius R is normalized to unit energy
    R: float
        Sphere radius that separates inner/outer components

    Returns
    -------
    B: Nx3 array
        Magnetic field produced by the sph components
    """

    basis = basis_fields(p, lmax, normalization=normalization, R=R)

    return basis[0] @ acoeffs + basis[1] @ bcoeffs


def basis_fields(p, lmax, normalization="default", R=1):
    """
    Computes magnetic field of each sph coefficient.
    Ignores the 'DC' component l=0.

    Parameters
    ----------
    p: Nx3 array
        coordinates in which the field is computed
    lmax: int
        maximum degree l used in the computation
    normalization: string
        "default": the fields correspond to magnetic normalized potential.
        "unit": the fields are normalized w.r.t integration over the unit sphere.
        "energy": each term in inner/outer basis function with respect to
            sphere with radius R is normalized to unit energy
    R: float
        Sphere radius that separates inner/outer components

    Returns
    -------
    B1: N x 3 x N_lmax array
        magnetic field at p for each alpha_lm
    B2: N x 3 x N_lmax array
        magnetic field at p for each beta_lm

    """
    mu0 = 1e-7 * 4 * np.pi
    L = lmax * (lmax + 2)  # Fixed
    B1 = np.zeros((L, p.shape[0], p.shape[1]))
    B2 = np.zeros((L, p.shape[0], p.shape[1]))

    sp = cartesian2spherical(p)

    idx = 0
    for l in range(1, lmax + 1):
        for m in range(-1 * l, l + 1):
            _Wlm = Wlm(l, m, sp[:, 1], sp[:, 2])
            _Wlm[:, 0] *= sp[:, 0] ** (l - 1)
            _Wlm[:, 1] *= sp[:, 0] ** (l - 1)
            _Wlm[:, 2] *= sp[:, 0] ** (l - 1)

            if normalization == "default":
                _Wlm *= np.sqrt(2 * l**2 + l) * mu0
            if normalization == "energy":
                _Wlm *= np.sqrt(2 * l**2 + l) * mu0
                _Wlm *= 1 / np.sqrt(R ** (2 * l + 1) * l * mu0)

            _Wlm = sphvec2cart(sp, _Wlm)
            B2[idx] = -1 * _Wlm  # r**l functions

            _Vlm = Vlm(l, m, sp[:, 1], sp[:, 2])
            if normalization == "default":
                _Vlm *= np.sqrt((2 * l + 1) * (l + 1)) * mu0
            if normalization == "energy":
                _Vlm *= np.sqrt((2 * l + 1) * (l + 1)) * mu0
                _Vlm *= np.sqrt(R ** (2 * l + 1) / ((l + 1) * mu0))

            _Vlm[:, 0] *= sp[:, 0] ** (-l - 2)
            _Vlm[:, 1] *= sp[:, 0] ** (-l - 2)
            _Vlm[:, 2] *= sp[:, 0] ** (-l - 2)
            _Vlm = sphvec2cart(sp, _Vlm)
            B1[idx] = -1 * _Vlm  # 1/r**l functions

            idx += 1

    # FIX, should be handled earlier maybe
    B1[np.isinf(B1)] = 0
    B2[np.isinf(B2)] = 0

    return np.moveaxis(B1, 0, 2), np.moveaxis(B2, 0, 2)


def compute_sphcoeffs_mesh(mesh, lmax, normalization="default", R=1):
    """
    Computes multipole moment (spherical harmonics coefficient) transformation
    from the mesh.

    Parameters
    ----------
    mesh: mesh object - the surface mesh
    lmax: int
        maximum degree l of the fit
    normalization: str
        'default' (Ylm**2 integrated over solid angle to 1)
        'energy' (field energy of basis fields normalized to 1 in R-ball)
    R: float
        radius in the energy normalization
    Returns
    -------
    Calm: (lmax*(lmax+2)xNvertices array
          transformation from the mesh to alm coefficients (r**l-terms)
    Cblm: (lmax*(lmax+2)xNvertices array
          transformation from the mesh to blm coefficients (r**(-l)-terms)

    """
    mu0 = 1e-7 * 4 * np.pi

    Gx, Gy, Gz = gradient_matrix(mesh, rotated=True)
    tri_normals, tri_areas = tri_normals_and_areas(mesh.vertices, mesh.faces)

    centers = np.mean(mesh.vertices[mesh.faces], axis=1)
    centers_sp = cartesian2spherical(centers)

    Calm = np.zeros((lmax * (lmax + 2), mesh.vertices.shape[0]))
    Cblm = np.zeros((lmax * (lmax + 2), mesh.vertices.shape[0]))

    idx = 0
    for l in range(1, lmax + 1):
        for m in range(-1 * l, l + 1):
            _Xlm = sphvec2cart(centers_sp, Xlm(l, m, *centers_sp[:, 1:].T))
            # Multiply by the normalization factor, and tri areas
            _Xlm *= np.sqrt(l * (l + 1)) * tri_areas[:, None]
            # alm ja blm have different r-dependencies
            alm_terms = _Xlm.T * (centers_sp[:, 0] ** (l))
            blm_terms = _Xlm.T * (centers_sp[:, 0] ** (-l - 1))

            # Combine rotated gradient (operates on stream function -> j)
            integral_alm = alm_terms[0] @ Gx + alm_terms[1] @ Gy + alm_terms[2] @ Gz
            integral_blm = blm_terms[0] @ Gx + blm_terms[1] @ Gy + blm_terms[2] @ Gz
            # Default normalization
            Cblm[idx] = -1 / ((2 * l + 1) * l) * integral_blm
            Calm[idx] = 1 / ((2 * l + 1) * (l + 1)) * integral_alm
            if normalization == "energy":
                Cblm[idx] *= np.sqrt(R ** (2 * l + 1) * l * mu0)
                Calm[idx] *= np.sqrt(((l + 1) * mu0) / (R ** (2 * l + 1)))

            idx += 1
        print("l = %d computed" % (l))

    return Calm, Cblm


###################################
# sph class


class SphBasis:
    """
    Constructs an object that describes unit sphere. Can be used to compute
    inner products and function sph spectra on the sphere.

    TODO: mu0 might be missing!!!

    Properties
    ----------
    sp: Nx3 array
        spherical coordinates of the points [r, theta, phi]
    p: Nx3 array
        cartesian coordinates
    qp: Mx3 array
        cartesian quadrature points
    sqp: Mx3 array
        spherical quadrature points
    """

    def __init__(self, Np):
        """
        Initialises the sphbasis object.

        Parameters
        ----------
        Np: int
            Mumber of points along theta and phi in spherical coordinates

        Returns
        -------
        self: sphbasis object

        """

        theta = np.linspace(0.01, np.pi - 0.01, Np)
        phi = np.linspace(0, 2 * np.pi, Np)
        phi, theta = np.meshgrid(phi, theta)
        phi = phi.flatten()
        theta = theta.flatten()

        self.sp = np.zeros((theta.shape[0], 3))
        self.sp[:, 0] = 1
        self.sp[:, 1] = theta
        self.sp[:, 2] = phi

        self.p = spherical2cartesian(self.sp)
        self.Np = Np

        self.initqp()

    def initqp(self):
        """
        Initialises quadrature points on the sphere.

        Default points are McLaren(10) so that we avoid singularities.
        """

        self.qp = get_spherical_quad_points()
        sp = cartesian2spherical(self.qp.points)
        self.sqp = sp

    def innerproduct(self, fun1, fun2):
        """
        Inner product of vector functions fun1 and fun2.
        Defined as integration over a surface of unit sphere <C,D> = int C dot D dOmega.
        Quadrature rule defined in qp is used.

        Parameters
        ----------
        fun1: Nx3 array
            vector function 1
        fun2: Nx3 array
            vector function 2

        Returns
        -------
        dotp: int
            inner product of fun1 and fun2

        """

        dotp = np.sum(self.qp.weights * np.sum(fun1 * fun2, axis=1)) * 4 * np.pi
        return dotp

    def avsphspectra(self, fun, lmax):
        """
        Calculate the l,m-spectra (over r**l-terms) of vector function defined in quadrature points
        using the inner product.

        Parameters
        ----------
        fun: Nx3 array
            vector function computed at quadrature points self.sqp
        lmax: int
            maximum degree l for which the spectra is computed

        Returns
        -------
        coeffs: lmax*(lmax+2)x1 arrays
            spectral coefficients

        """

        coeffs = []

        for l in range(1, lmax + 1):
            for m in range(-1 * l, l + 1):
                _Wlm = Wlm(l, m, self.sqp[:, 1], self.sqp[:, 2])
                ctemp = self.innerproduct(fun, _Wlm)
                ctemp /= self.sqp[0, 0] ** (l - 1) * np.sqrt(
                    2 * l**2 + l
                )  # we use this normalization
                #                ctemp /= (self.sqp[0,0]**(l-1))
                coeffs.append(ctemp)

        coeffs = np.array(coeffs)
        return coeffs

    def bvsphspectra(self, fun, lmax):
        """
        Calculate the l,m-spectra (over r**(-l)-terms) of vector function defined in quadrature points
        using the inner product.

        Parameters
        ----------
        fun: Nx3 array
            vector function computed at quadrature points self.sqp
        lmax: int
            maximum degree l for which the spectra is computed

        Returns
        -------
        coeffs: lmax*(lmax+2)x1 arrays
            spectral coefficients

        """

        coeffs = []

        for l in range(1, lmax + 1):
            for m in range(-1 * l, l + 1):
                _Vlm = Vlm(l, m, self.sqp[:, 1], self.sqp[:, 2])
                ctemp = self.innerproduct(fun, _Vlm)
                ctemp /= self.sqp[0, 0] ** (l - 1) * np.sqrt(
                    (l + 1) * (2 * l + 1)
                )  # we use this normalization
                #                ctemp /= (self.sqp[0,0]**(l-1))
                coeffs.append(ctemp)

        coeffs = np.array(coeffs)
        return coeffs


###################################
# Functions for plotting spherical harmonics


def plotYlms(sph, lmax, polar=False):
    """
    Plots real spherical harmonics up to lmax.
    Inspired by https://docs.enthought.com/mayavi/mayavi/auto/example_spherical_harmonics.html.

    Parameters
    ----------
    sph: spherical harmonics analysis object
    lmax: int

        maximum degree l
    polar: boolean
        plot polar representation?
    """
    from mayavi import mlab

    theta = np.reshape(sph.sp[:, 1], (sph.Np, sph.Np))
    phi = np.reshape(sph.sp[:, 2], (sph.Np, sph.Np))
    r = 0.4
    x = r * np.sin(theta) * np.cos(phi)
    y = r * np.sin(theta) * np.sin(phi)
    z = r * np.cos(theta)

    if polar:
        for l in range(1, lmax + 1):
            for m in range(l):
                _ylm = ylm(l, m, theta.flatten(), phi.flatten())
                _ylm = np.reshape(_ylm, (sph.Np, sph.Np))

                mlab.mesh(x - m, y - l, z, scalars=_ylm, colormap="bwr")
                _ylm /= _ylm.max()
                mlab.mesh(
                    _ylm * x - m,
                    _ylm * y - l,
                    _ylm * z + 1.3,
                    scalars=np.abs(_ylm),
                    colormap="Spectral",
                )

        mlab.view(90, 70, 6.2, (-1.3, -2.9, 0.25))
    else:
        for l in range(0, lmax + 1):
            for m in range(-l, l + 1):
                _ylm = ylm(l, m, theta.flatten(), phi.flatten())
                _ylm = np.reshape(_ylm, (sph.Np, sph.Np))

                mlab.mesh(x - m, y - l, z, scalars=_ylm, colormap="bwr")

        mlab.view(0, 180)


def plotYlm(sph, l, m):
    """
    Plots real spherical harmonics of order m and degree l.
    Inspired by https://docs.enthought.com/mayavi/mayavi/auto/example_spherical_harmonics.html

    Parameters
    ----------
    sph: spherical harmonics analysis object
    l: int
        degree l
    m: int
        order m

    """
    from mayavi import mlab

    theta = np.reshape(sph.sp[:, 1], (sph.Np, sph.Np))
    phi = np.reshape(sph.sp[:, 2], (sph.Np, sph.Np))
    r = 0.6
    x = r * np.sin(theta) * np.cos(phi)
    y = r * np.sin(theta) * np.sin(phi)
    z = r * np.cos(theta)

    _ylm = ylm(l, m, theta.flatten(), phi.flatten())
    _ylm = np.reshape(_ylm, (sph.Np, sph.Np))

    mlab.mesh(x - m, y - l, z, scalars=_ylm, colormap="bwr")

    _ylm /= _ylm.max()
    mlab.mesh(
        _ylm * x - m,
        _ylm * y - l,
        _ylm * z + 1.3,
        scalars=np.abs(_ylm),
        colormap="Spectral",
    )


def plotWlm(sph, l, m):
    """
    Plots magnetic field basis function 'Wlm' (r**l) over a sphere.

    Parameters
    ----------
    sph: spherical harmonics analysis object
    l: int
        degree l
    m: int
        order m

    Returns
    -------
    obj: mayavi object

    """
    from mayavi import mlab

    _Wlm = Wlm(l, m, sph.sp[:, 1], sph.sp[:, 2])
    _Wlm = sphvec2cart(sph.sp, _Wlm)
    obj = mlab.quiver3d(
        sph.p[:, 0], sph.p[:, 1], sph.p[:, 2], _Wlm[:, 0], _Wlm[:, 1], _Wlm[:, 2]
    )
    obj.glyph.glyph_source.glyph_source.center = np.array((0, 0, 0))
    return obj


def plotBWlm_volume(sph, l, m, lim, Np, offset):
    """
    Plots magnetic field basis function 'Wlm' (r**l) over a 3D volume.

    Parameters
    ----------
    sph: spherical harmonics analysis object
    l: int
        degree l
    m: int
        order m
    lim: float
        limits for coordinates, e.g., xmin = -lim, xmax = lim
    Np: int
        number of points along different coordinates
    offset: 1x3 array
        offset of the volume in which Wlm is plotted

    Returns
    -------
        obj: mayavi object

    """

    from mayavi import mlab

    x, y, z = np.meshgrid(
        np.linspace(-lim + offset[0], lim + offset[0], Np),
        np.linspace(-lim + offset[1], lim + offset[1], Np),
        np.linspace(-lim + offset[2], lim + offset[2], Np),
    )

    p = np.array((x.flatten(), y.flatten(), z.flatten())).T
    sp = cartesian2spherical(p)

    _Wlm = Wlm(l, m, sp[:, 1], sp[:, 2])
    _Wlm *= np.sqrt(2 * l**2 + l)
    _Wlm[:, 0] *= sp[:, 0] ** (l - 1)
    _Wlm[:, 1] *= sp[:, 0] ** (l - 1)
    _Wlm[:, 2] *= sp[:, 0] ** (l - 1)

    _Wlm = sphvec2cart(sp, _Wlm)
    obj = mlab.quiver3d(p[:, 0], p[:, 1], p[:, 2], _Wlm[:, 0], _Wlm[:, 1], _Wlm[:, 2])
    obj.glyph.glyph_source.glyph_source.center = np.array((0, 0, 0))
    return obj


def plotVlm(sph, l, m):
    """
    Plots magnetic field basis function 'Vlm' (r**(-l)) over a sphere.

    Parameters
    ----------
    sph: spherical harmonics analysis object
    l: int
        degree l
    m: int
        order m

    Returns
    -------
    obj: mayavi object

    """
    from mayavi import mlab

    _Vlm = Vlm(l, m, sph.sp[:, 1], sph.sp[:, 2])
    _Vlm = sphvec2cart(sph.sp, _Vlm)
    obj = mlab.quiver3d(
        sph.p[:, 0], sph.p[:, 1], sph.p[:, 2], _Vlm[:, 0], _Vlm[:, 1], _Vlm[:, 2]
    )
    obj.glyph.glyph_source.glyph_source.center = np.array((0, 0, 0))
    return obj


def plotBVlm_volume(sph, l, m, lim, Np, offset):
    """
    Plots magnetic field basis function 'Vlm' (r**(-l)) over a 3D volume.

    Parameters
    ----------
    sph: spherical harmonics analysis object
    l: int
        degree l
    m: int
        order m
    lim: float
        limits for coordinates, e.g., xmin = -lim, xmax = lim
    Np: int
        number of points along different coordinates
    offset: 1x3 array
        offset of the volume in which Vlm is plotted

    Returns
    -------
    obj: mayavi object

    """
    from mayavi import mlab

    x, y, z = np.meshgrid(
        np.linspace(-lim + offset[0], lim + offset[0], Np),
        np.linspace(-lim + offset[1], lim + offset[1], Np),
        np.linspace(-lim + offset[2], lim + offset[2], Np),
    )

    p = np.array((x.flatten(), y.flatten(), z.flatten())).T
    sp = cartesian2spherical(p)

    _Vlm = Vlm(l, m, sp[:, 1], sp[:, 2])
    _Vlm *= np.sqrt((l + 1) * (2 * l + 1))

    _Vlm[:, 0] *= sp[:, 0] ** (-1 * (l + 2))
    _Vlm[:, 1] *= sp[:, 0] ** (-1 * (l + 2))
    _Vlm[:, 2] *= sp[:, 0] ** (-1 * (l + 2))

    _Vlm = sphvec2cart(sp, _Vlm)
    obj = mlab.quiver3d(p[:, 0], p[:, 1], p[:, 2], _Vlm[:, 0], _Vlm[:, 1], _Vlm[:, 2])
    obj.glyph.glyph_source.glyph_source.center = np.array((0, 0, 0))
    return obj


def plotXlm(sph, l, m):
    """
    Plots vector spherical harmonic basis function 'Xlm' over a sphere.

    Parameters
    ----------
    sph: spherical harmonics analysis object
    l: int
        degree l
    m: int
        order m

    Returns
    -------
    obj: mayavi object

    """
    from mayavi import mlab

    _Xlm = Xlm(l, m, sph.sp[:, 1], sph.sp[:, 2])
    _Xlm = sphvec2cart(sph.sp, _Xlm)
    obj = mlab.quiver3d(
        sph.p[:, 0], sph.p[:, 1], sph.p[:, 2], _Xlm[:, 0], _Xlm[:, 1], _Xlm[:, 2]
    )
    obj.glyph.glyph_source.glyph_source.center = np.array((0, 0, 0))
    return obj



from . import mesh_conductor
from . import mesh_impedance
from . import mesh_magnetics
from . import mesh_calculus
from . import line_magnetics
from . import line_conductor
from . import coil_optimize
from . import sphtools
from . import suhtools
from . import thermal_noise
from . import utils
from . import integrals
from . import legacy
from . import flatten_mesh

try:
    from .version import __version__
except ModuleNotFoundError:
    try:
        print("version.py not present, did you install the package?")
        print("Attempting to get version through pkg_resources")
        from pkg_resources import get_distribution

        __version__ = get_distribution("bfieldtools").version
    except ModuleNotFoundError:
        print("pkg_resources not found, I'm giving up")
        print("Setting version to 'unknown'")
        __version__ = "unknown"


__all__ = [
    "line_magnetics",
    "line_conductor",
    "coil_optimize",
    "mesh_calculus",
    "mesh_magnetics",
    "mesh_conductor",
    "mesh_impedance",
    "sphtools",
    "suhtools",
    "thermal_noise",
    "utils",
    "integrals",
    "legacy",
    "flatten_mesh",
]



"""
Test diagonal of inductance matrix
===================================================

Use different number of quadrature points and two different implementations
"""

import numpy as np
import matplotlib.pyplot as plt

from bfieldtools.utils import load_example_mesh

coilmesh = load_example_mesh("10x10_plane")

from bfieldtools.mesh_impedance import self_inductance_matrix, mutual_inductance_matrix


M5 = mutual_inductance_matrix(coilmesh, coilmesh, quad_degree=5)
M6 = mutual_inductance_matrix(coilmesh, coilmesh, quad_degree=6)
M7 = mutual_inductance_matrix(coilmesh, coilmesh, quad_degree=7)

plt.figure()
for m in (M5, M6, M7):
    plt.plot(np.diag(m))

MM5 = self_inductance_matrix(coilmesh, quad_degree=5, analytic_self_coupling=True)
MM6 = self_inductance_matrix(coilmesh, quad_degree=6, analytic_self_coupling=True)
MM7 = self_inductance_matrix(coilmesh, quad_degree=7, analytic_self_coupling=True)

#%% Plot the diagonals

plt.gca().set_prop_cycle(None)
for m in (MM5, MM6, MM7):
    plt.plot(np.diag(m), "--")



"""
Example of designing a shielded biplanar coil
===============================================
"""

import numpy as np
import matplotlib.pyplot as plt
from mayavi import mlab
import trimesh

from bfieldtools.mesh_conductor import MeshConductor, StreamFunction
from bfieldtools.contour import scalar_contour
from bfieldtools.viz import plot_3d_current_loops
from bfieldtools.utils import load_example_mesh, combine_meshes


# Set unit, e.g. meter or millimeter.
# This doesn't matter, the problem is scale-invariant
scaling_factor = 0.1

# Load simple plane mesh that is centered on the origin
planemesh = load_example_mesh("10x10_plane_hires")
planemesh.apply_scale(scaling_factor)

# Specify coil plane geometry
center_offset = np.array([0, 0, 0]) * scaling_factor
standoff = np.array([0, 4, 0]) * scaling_factor

# Create coil plane pairs
coil_plus = trimesh.Trimesh(
    planemesh.vertices + center_offset + standoff, planemesh.faces, process=False
)

coil_minus = trimesh.Trimesh(
    planemesh.vertices + center_offset - standoff, planemesh.faces, process=False
)

mesh1 = combine_meshes((coil_minus, coil_plus))
mesh2 = mesh1.copy()
mesh2.apply_scale(1.4)

coil1 = MeshConductor(mesh_obj=mesh1, basis_name="inner", N_sph=4)
coil2 = MeshConductor(mesh_obj=mesh2, basis_name="inner", N_sph=4)

M11 = coil1.inductance
M22 = coil2.inductance
M21 = coil2.mutual_inductance(coil1)
# Mapping from I1 to I2, constraining flux through mesh2 to zero
P = -np.linalg.solve(M22, M21)

A1, Beta1 = coil1.sph_couplings
A2, Beta2 = coil2.sph_couplings

# Use lines below to get coulings with different normalization
# from bfieldtools.sphtools import compute_sphcoeffs_mesh
# A1, Beta1 = compute_sphcoeffs_mesh(mesh1, 5, normalization='energy', R=1)
# A2, Beta2 = compute_sphcoeffs_mesh(mesh2, 5, normalization='energy', R=1)

# Beta1 = Beta1[:, coil1.inner_vertices]
# Beta2 = Beta2[:, coil2.inner_vertices]

x = y = np.linspace(-0.8, 0.8, 50)  # 150)
X, Y = np.meshgrid(x, y, indexing="ij")
points = np.zeros((X.flatten().shape[0], 3))
points[:, 0] = X.flatten()
points[:, 1] = Y.flatten()


CB1 = coil1.B_coupling(points)
CB2 = coil2.B_coupling(points)

CU1 = coil1.U_coupling(points)
CU2 = coil2.U_coupling(points)


#%% Precalculations for the solution
# alpha[15] = 1
# Minimization of magnetic energy with spherical harmonic constraint
C = Beta1 + Beta2 @ P
M = M11 + M21.T @ P

from scipy.linalg import eigvalsh

ssmax = eigvalsh(C.T @ C, M, eigvals=[M.shape[1] - 1, M.shape[1] - 1])

#%% Specify spherical harmonic and calculate corresponding shielded field
beta = np.zeros(Beta1.shape[0])
beta[7] = 1  # Gradient
# beta[2] = 1  # Homogeneous

# Minimum residual
_lambda = 1e3
# Minimum energy
# _lambda=1e-3
I1inner = np.linalg.solve(C.T @ C + M * ssmax / _lambda, C.T @ beta)

I2inner = P @ I1inner

s1 = StreamFunction(I1inner, coil1)
s2 = StreamFunction(I2inner, coil2)

# s = mlab.triangular_mesh(*mesh1.vertices.T, mesh1.faces, scalars=I1)
# s.enable_contours=True
# s = mlab.triangular_mesh(*mesh2.vertices.T, mesh2.faces, scalars=I2)
# s.enable_contours=True

B1 = CB1 @ s1
B2 = CB2 @ s2

U1 = CU1 @ s1
U2 = CU2 @ s2
#%% Plot
cc1 = scalar_contour(mesh1, mesh1.vertices[:, 2], contours=[-0.001])
cc2 = scalar_contour(mesh2, mesh2.vertices[:, 2], contours=[-0.001])
cx10 = cc1[0][:, 1]
cy10 = cc1[0][:, 0]
cx20 = cc2[0][:, 1]
cy20 = cc2[0][:, 0]

cx11 = cc1[1][:, 1]
cy11 = cc1[1][:, 0]
cx21 = cc2[1][:, 1]
cy21 = cc2[1][:, 0]

B = (B1.T + B2.T)[:2].reshape(2, x.shape[0], y.shape[0])
lw = np.sqrt(B[0] ** 2 + B[1] ** 2)
lw = 2 * np.log(lw / np.max(lw) * np.e + 1.1)

xx = np.linspace(-1, 1, 16)
# seed_points = 0.56*np.array([xx, -np.sqrt(1-xx**2)])
# seed_points = np.hstack([seed_points, (0.56*np.array([xx, np.sqrt(1-xx**2)]))])
# seed_points = np.hstack([seed_points, (0.56*np.array([np.zeros_like(xx), xx]))])
seed_points = np.array([cx10 + 0.001, cy10])
seed_points = np.hstack([seed_points, np.array([cx11 - 0.001, cy11])])
seed_points = np.hstack([seed_points, (0.56 * np.array([np.zeros_like(xx), xx]))])

# plt.streamplot(x,y, B[1], B[0], density=2, linewidth=lw, color='k',
#               start_points=seed_points.T, integration_direction='both')
U = (U1 + U2).reshape(x.shape[0], y.shape[0])
U /= np.max(U)
plt.figure()
plt.contourf(X, Y, U.T, cmap="seismic", levels=40)
# plt.imshow(U, vmin=-1.0, vmax=1.0, cmap='seismic', interpolation='bicubic',
#           extent=(x.min(), x.max(), y.min(), y.max()))
plt.streamplot(
    x,
    y,
    B[1],
    B[0],
    density=2,
    linewidth=lw,
    color="k",
    start_points=seed_points.T,
    integration_direction="both",
    arrowsize=0.1,
)

# plt.plot(seed_points[0], seed_points[1], '*')

plt.plot(cx10, cy10, linewidth=3.0, color="gray")
plt.plot(cx20, cy20, linewidth=3.0, color="gray")
plt.plot(cx11, cy11, linewidth=3.0, color="gray")
plt.plot(cx21, cy21, linewidth=3.0, color="gray")
plt.axis("image")

plt.xticks([])
plt.yticks([])

#%%
N = 20
mm = max(abs(s1))
dd = 2 * mm / N
vmin = -dd * N / 2 + dd / 2
vmax = dd * N / 2 - dd / 2
contour_vals1 = np.arange(vmin, vmax, dd)
mm = max(abs(s2))
N2 = (2 * mm - dd) // dd
if N2 % 2 == 0:
    N2 -= 1
vmin = -dd * N2 / 2
vmax = mm
contour_vals2 = np.arange(vmin, vmax, dd)
contours1 = scalar_contour(mesh1, s1.vert, contours=contour_vals1)
contours2 = scalar_contour(mesh2, s2.vert, contours=contour_vals2)


def setscene(scene1, coil):
    scene1.actor.mapper.interpolate_scalars_before_mapping = True
    scene1.module_manager.scalar_lut_manager.number_of_colors = 32
    scene1.scene.y_plus_view()
    if coil == 1:
        scene1.scene.camera.position = [
            4.7267030067743576e-08,
            2.660205137153174,
            8.52196480605194e-08,
        ]
        scene1.scene.camera.focal_point = [
            4.7267030067743576e-08,
            0.4000000059604645,
            8.52196480605194e-08,
        ]
        scene1.scene.camera.view_angle = 30.0
        scene1.scene.camera.view_up = [1.0, 0.0, 0.0]
        scene1.scene.camera.clipping_range = [1.116284842928313, 2.4468228732691104]
        scene1.scene.camera.compute_view_plane_normal()
    else:
        scene1.scene.camera.position = [
            4.7267030067743576e-08,
            3.7091663385397116,
            8.52196480605194e-08,
        ]
        scene1.scene.camera.focal_point = [
            4.7267030067743576e-08,
            0.4000000059604645,
            8.52196480605194e-08,
        ]
        scene1.scene.camera.view_angle = 30.0
        scene1.scene.camera.view_up = [1.0, 0.0, 0.0]
        scene1.scene.camera.clipping_range = [2.948955346473114, 3.40878670176758]
        scene1.scene.camera.compute_view_plane_normal()
    scene1.scene.render()
    scene1.scene.anti_aliasing_frames = 20
    scene1.scene.magnification = 2


fig = mlab.figure(bgcolor=(1, 1, 1), size=(400, 400))
fig = plot_3d_current_loops(
    contours1, tube_radius=0.005, colors=(0.9, 0.9, 0.9), figure=fig
)
m = abs(s1).max()
mask = mesh1.triangles_center[:, 1] > 0
faces1 = mesh1.faces[mask]
surf = mlab.triangular_mesh(
    *mesh1.vertices.T, faces1, scalars=s1.vert, vmin=-m, vmax=m, colormap="seismic"
)
setscene(surf, 1)

fig = mlab.figure(bgcolor=(1, 1, 1), size=(400, 400))
fig = plot_3d_current_loops(
    contours2, tube_radius=0.005, colors=(0.9, 0.9, 0.9), figure=fig
)
faces2 = mesh2.faces[mesh2.triangles_center[:, 1] > 0]
surf = mlab.triangular_mesh(
    *mesh2.vertices.T, faces2, scalars=s2.vert, vmin=-m, vmax=m, colormap="seismic"
)
setscene(surf, 2)


#%% Plot the coil surface and the field plane
fig = mlab.figure(bgcolor=(1, 1, 1))
surf = mlab.triangular_mesh(*mesh1.vertices.T, mesh1.faces, color=(0.8, 0.2, 0.2))
surf.actor.property.edge_visibility = True
surf.actor.property.render_lines_as_tubes = True
surf.actor.property.line_width = 1.2


surf = mlab.triangular_mesh(*mesh2.vertices.T, mesh2.faces, color=(0.2, 0.2, 0.8))
surf.actor.property.edge_visibility = True
surf.actor.property.render_lines_as_tubes = True
surf.actor.property.line_width = 1.2
# Plot plane
plane = mlab.triangular_mesh(
    np.array([x[0], x[-1], x[-1], x[0]]),
    np.array([x[0], x[0], x[-1], x[-1]]),
    np.zeros(4),
    np.array([[0, 1, 2], [2, 3, 0]]),
    color=(0.7, 0.7, 0.7),
    opacity=0.7,
)



"""
functions for flattening a mesh with a boundary
"""

__all__ = [
    "A_matrix_complex",
    "eigen_complex_laplacian",
    "flatten_mesh",
    "mesh2plane",
    "plane2mesh",
]

import numpy as np
from bfieldtools.mesh_calculus import laplacian_matrix, mass_matrix
from scipy.sparse.linalg import eigsh
from scipy.sparse import csc_matrix
import trimesh


def A_matrix_complex(mesh):
    """
    Area matrix for complex parametrization
    See this course:
    https://www.cs.cmu.edu/~kmcrane/Projects/DDG/
    section 7.4.

    CURRENTLY ONLY FOR A MESH WITH A SINGLE BOUNDARY (no holes)

    Parameters
    ----------
    mesh : trimesh

    Returns
    -------
    A : area matrix
        u.T.conj() @ A @ u is the area of the mesh
        where u is the complex coordinates of the vertices
        (complex parameterization)

    """
    p = mesh.outline().entities[0]
    b = p.points
    N = mesh.vertices.shape[0]
    ii = np.hstack([b, np.roll(b, 1)])
    jj = np.hstack([np.roll(b, 1), b])
    vv = np.ones(len(b))
    data = -0.25 * 1j * np.hstack([vv, -vv])
    A = csc_matrix((data, (ii, jj)), shape=(N, N))

    return A


def eigen_complex_laplacian(mesh, Nc, _lambda):
    L = laplacian_matrix(mesh)
    M = mass_matrix(mesh)
    Ac = A_matrix_complex(mesh)
    vals, uv = eigsh(-0.5 * L.T - _lambda * Ac, Nc, M, which="LM", sigma=0)
    return vals, uv


def flatten_mesh(mesh, _lambda=1.0):
    """Flatten the mesh, return uv coordinates and the mesh in 2D

    Parameters
    ----------
    mesh : Trimesh object
        must have boundary
    _lambda : int <= 1.0
        parameter for trading of area-distortion/angle-preservation.
        The default is 1.0

    Returns
    -------
    u : array
        first coordinate of the paramterization
    v : array
        second coordinate of the paramterization
    mesh2d : Trimesh object with coordinates (u,v,0)

    _lambda <= 1.0
    _lambda == 1.0 => conformal mapping
    _lambda == 0.5 =>  not conformal but less area distortion
    _lambda --> 0 mapping becomes denegerate (real==imag)
    _lambda > 1 (e.g. 1.01-1.1) folding effects
    """
    vals, uv = eigen_complex_laplacian(mesh, 2, _lambda)

    # Coordinates with initial phase
    u = uv[:, 1].real
    v = uv[:, 1].imag

    # Determine "phase" by matching the uv coordinate function with mesh coordinates
    theta = np.linspace(0, 2 * np.pi, 50)
    yy = np.imag(np.exp(1j * theta)[:, None] * uv[:, 1])
    # plt.plot(np.sum(mesh.vertices[:,0]*xx, axis=1))
    ii = np.argmax(np.sum(mesh.vertices[:, 1] * yy, axis=1))

    theta = theta[ii]
    u = np.real(np.exp(1j * theta) * uv[:, 1])
    v = np.imag(np.exp(1j * theta) * uv[:, 1])

    mesh2d = trimesh.Trimesh(np.array([u, v, 0 * u]).T, mesh.faces, process=False)
    return u, v, mesh2d


# Map points from 2D to 3D or vice versa
def mesh2plane(points3d, mesh, u, v):
    """
    Map points from 3D to 2D on the u,v -plane

    Parameters
    ----------
    points3d : nparray
        points close to the  mesh (N, 3)
    mesh : Trimesh object
        3D mesh with the u,v -parametrisation
    u : ndarray
        u coordinate (N,)
    v : ndarray
        v coordinate (N,)

    Returns
    -------
    ndarray
       uv-coordinates of the 3D points

    """
    c, d, f = trimesh.proximity.closest_point(mesh, points3d)
    tris = mesh.vertices[mesh.faces[f]]
    barys = trimesh.triangles.points_to_barycentric(tris, c)
    print(barys)
    p1 = np.sum(u[mesh.faces[f]] * barys, axis=1)
    p2 = np.sum(v[mesh.faces[f]] * barys, axis=1)
    return np.array([p1, p2]).T


def plane2mesh(points2d, mesh, u, v):
    """
    Map point from 2D to 3D on the mesh

    Parameters
    ----------
    points3d : nparray
        points close to the  mesh (N, 3)
    mesh : Trimesh object
        3D mesh with the u,v -parametrisation
    u : ndarray
        u coordinate (N,)
    v : ndarray
        v coordinate (N,)

    Returns
    -------
    ndarray
       uv-coordinates of the 3D points
    """
    mesh2d = trimesh.Trimesh(np.array([u, v, 0 * u]).T, mesh.faces)
    c, d, f = trimesh.proximity.closest_point(mesh2d, points2d)
    # Leave points outside the mesh area
    c = c[d < 1e-8]
    f = f[d < 1e-8]
    # Homogeneous coordinates
    c[:, 2] = 1
    p = []
    for ci, fi in zip(c, f):
        R = np.ones((3, 3))
        R[0] = u[mesh.faces[fi]]
        R[1] = v[mesh.faces[fi]]
        bary = np.linalg.solve(R, ci)
        p.append(mesh.vertices[mesh.faces[fi], :].T @ bary)

    return np.array(p)



"""
Solving linearly polarizing material in a homogenenous field
------------------------------------------------------------

The example is based on the following integral equation of 
a piecewise harmonic potential :math:`U` on the boundary of the polarizing material

.. math::
    \\frac{\\mu_0 + \\mu}{2} U(\\vec{r}) = \\mu_0 U_\\infty(\\vec{r}) + \\frac{\\mu_0 - \\mu}{4\\pi}\\int_{S} U(\\vec{r}\\,') (\\hat{n}' \\cdot \\nabla') \\frac{1}{|\\vec{r}-\\vec{r}\\,'|} dS'\\,.
                                    
* :math:`U` is the magnetic scalar potential to be solved
* :math:`\\mu_0` is the permeability of the free space
* :math:`\\mu` is the permeability of the material
* :math:`U_\\infty` is the magnetic scalar potential without the material
* :math:`S` is the surface of the material

After discretization :math:`U(\\vec{r})=\sum_n u_n h_n(\\vec{r})` and integrating the integral equation over every hat :math:`h_n(\\vec{r})`,
a linear system of equations is obtained

.. math::
    \\mathbf{N}\\mathbf{u} = \\frac{2}{1+\\mu_\\mathrm{r}} \\mathbf{u}_\\infty + \\frac{1 - \\mu_\\mathrm{r}}{4\\pi(1 + \\mu_\\mathrm{r})}\\mathbf{D}\\mathbf{u}\,.

* :math:`\\mathbf{u}` is :math:`U` at the mesh nodes
* :math:`\\mathbf{u}_\\infty` is the potential :math:`U_\\infty` integrated over hat functions
* :math:`\\mu_\\mathrm{r}` is the relative permeability of the material
* :math:`\\mathbf{D}` is the double-layer coupling matrix
* :math:`\\mathbf{N}` is the mesh mass matrix
"""


import numpy as np
import matplotlib.pyplot as plt
from mayavi import mlab

from bfieldtools.mesh_conductor import MeshConductor as Conductor
from bfieldtools.mesh_magnetics import magnetic_field_coupling
from bfieldtools.mesh_magnetics import scalar_potential_coupling
from bfieldtools.mesh_calculus import mass_matrix
from bfieldtools.utils import load_example_mesh


from trimesh.creation import icosphere

# Use a sphere
# mesh = icosphere(3, 1)

# Use a cube
mesh = load_example_mesh("cube")
mesh.vertices -= mesh.vertices.mean(axis=0)
mesh.apply_scale(0.2)


def Dmatrix(mesh1, mesh2, Nchunks=100):
    """
    "Double-layer potential" coupling between two meshes
    using a Galerkin method with hat basis

    Discretize integral equations using hat functions
    on on both meshes. Potential from mesh1 hat functions
    is calculated analytically and integrated over the
    hat functions of mesh2 numerically

    Parameters
    ----------
    mesh1 : Trimesh object
    mesh2 : Trimesh object
    Nchunks : int, optional
        Number of chunks in the potential calculation. The default is 100.

    Returns
    -------
    None.

    """
    face_points = mesh2.vertices[mesh2.faces]
    weights = np.array([[0.5, 0.25, 0.25], [0.25, 0.5, 0.25], [0.25, 0.25, 0.5]])
    # Combine vertices for quadrature points
    Rquad = np.einsum("...ij,ik->...kj", face_points, weights)
    R = Rquad.reshape(-1, 3)
    U = scalar_potential_coupling(
        mesh1, R, Nchunks, multiply_coeff=False, approx_far=True
    )

    face_areas = mesh2.area_faces
    # Reshape and multiply by quadrature weights
    Dcomps = U.reshape(Rquad.shape[:2] + (len(mesh2.vertices),)) * (
        face_areas[:, None, None] / 3
    )
    # Sum up the quadrature points
    D = mesh2.faces_sparse @ Dcomps[:, 0, :]
    D += mesh2.faces_sparse @ Dcomps[:, 1, :]
    D += mesh2.faces_sparse @ Dcomps[:, 2, :]

    if mesh1 is mesh2:
        # Recalculate diagonals
        d = np.diag(D)
        D -= np.diag(d)
        # Make rows sum to -2*pi*(vertex area), should be more accurate
        d2 = -2 * np.pi * mass_matrix(mesh2, lumped=False) - np.diag(D.sum(axis=1))
        D += d2
        # Make D solvable by adding rank-1 matrix
        D += np.ones_like(D) * np.max(np.linalg.svd(D, False, False)) / D.shape[1]

    return D


# Some linear input potentials -> uniform field
def phi0x(r):
    return r[:, 0]


def phi0y(r):
    return r[:, 1]


def phi0z(r):
    return r[:, 2]


def project_to_hats(mesh, func):
    """
    Numerically integrate func over hat functions

    Parameters
    ----------
    mesh : Trimesh object
        the domain for hat functions
    func : function
        potential function for phi0, takes (N,3) array
        of points as input

    Returns
    -------
    p_mat : (Nvertices,) array
        func projected on the hat functions

    """
    # Index vertex points for each faces
    face_points = mesh.vertices[mesh.faces]
    weights = np.array([[0.5, 0.25, 0.25], [0.25, 0.5, 0.25], [0.25, 0.25, 0.5]])
    # Combine vertices for quadrature points
    Rquad = np.einsum("...ij,ik->...kj", face_points, weights)
    R = Rquad.reshape(-1, 3)

    # Evaluation func at quadrature points
    p = func(R)
    face_areas = mesh.area_faces
    # Reshape and multiply by quadrature weights
    pcomps = p.reshape(Rquad.shape[:2]) * (face_areas[:, None] / 3)
    # Sum up the quadrature points
    p_mat = mesh.faces_sparse @ pcomps[:, 0]
    p_mat += mesh.faces_sparse @ pcomps[:, 1]
    p_mat += mesh.faces_sparse @ pcomps[:, 2]

    return p_mat


#%% Compute matrices
print("Computing D matrix")
D = Dmatrix(mesh, mesh)
print("Computing mass matrix")
M = mass_matrix(mesh)
Ml = mass_matrix(mesh, lumped=True)


#%% Solve linear BEM equations
print("Computing input potential")
pp = project_to_hats(mesh, phi0x)
mu_r = 100
c1 = (1 - mu_r) / (1 + mu_r)
u = np.linalg.solve(M - 1 / (2 * np.pi) * c1 * D, -2 * pp / (mu_r + 1))

# Plot potential on the mesh
mlab.figure("Potential on the boundary", bgcolor=(1, 1, 1))
m = mlab.triangular_mesh(*mesh.vertices.T, mesh.faces, scalars=u, colormap="bwr")
m.actor.mapper.interpolate_scalars_before_mapping = True
m.module_manager.scalar_lut_manager.number_of_colors = 32
#%% Compute fields on a plane
# Load plane for the visualization of the potential
plane = load_example_mesh("10x10_plane_hires", process=True)
t = np.eye(4)
t[1:3, 1:3] = np.array([[0, 1], [-1, 0]])
plane.apply_transform(t)
plane.apply_scale(0.3)
plane = plane.subdivide()

Uplane = scalar_potential_coupling(
    mesh, plane.vertices, multiply_coeff=False, approx_far=True
)

uprim = phi0x(plane.vertices)
usec = (mu_r - 1) / (4 * np.pi) * Uplane @ u
uplane = uprim + usec

# Meshgrid on the same plane for the bfield
X, Y = np.meshgrid(
    np.linspace(-1.5, 1.5, 50), np.linspace(-1.5, 1.5, 50), indexing="ij"
)
pp = np.zeros((50 * 50, 3))
pp[:, 0] = X.flatten()
pp[:, 1] = Y.flatten()

Bplane = magnetic_field_coupling(mesh, pp, analytic=True)
bprim = pp * 0  # copy pp
# add x directional field
mu0 = 1e-7 * 4 * np.pi
bprim[:, 0] = -1
# In this simulation mu0==1 is assumed
# magnetic_field_coupling uses mu0 in SI units
bplane = (mu_r - 1) / (4 * np.pi) / mu0 * Bplane @ u + bprim
# uplane[Uplane.sum(axis=1)]

#%% Plot the results
mlab.figure("Potential for B", bgcolor=(1, 1, 1), size=(1000, 1000))
m = mlab.triangular_mesh(*plane.vertices.T, plane.faces, scalars=uplane, colormap="bwr")
m.actor.mapper.interpolate_scalars_before_mapping = True
m.module_manager.scalar_lut_manager.number_of_colors = 32
# vectors = mlab.quiver3d(*(plane.triangles_center + np.array([0,0,0.001])).T,
#                         *(-gradient(uplane, plane)), color=(0,0,0),
#               scale_mode='none', scale_factor=0.01, mode='arrow')
# vectors.glyph.mask_input_points = True
# vectors.glyph.mask_points.random_mode_type = 0
# vectors.glyph.mask_points.on_ratio = 4
# vectors.glyph.glyph_source.glyph_position = 'center'
m.scene.z_plus_view()

mlab.figure("B field", bgcolor=(1, 1, 1), size=(1000, 1000))
vectors2 = mlab.quiver3d(
    *pp.T,
    *bplane.T,
    color=(1, 0, 0),
    scale_mode="vector",
    scale_factor=0.02,
    mode="arrow"
)
vectors2.glyph.glyph.scale_factor = 0.08
vectors2.glyph.mask_input_points = True
vectors2.glyph.mask_points.random_mode_type = 0
vectors2.glyph.mask_points.on_ratio = 4
vectors2.glyph.glyph_source.glyph_position = "center"
vectors2.scene.z_plus_view()

# Streamline plot for the cube
bplane[:, 2] = 0
vecfield = mlab.pipeline.vector_field(
    *pp.T.reshape(3, 50, 50, 1), *bplane.T.reshape(3, 50, 50, 1)
)
vecnorm = mlab.pipeline.extract_vector_norm(vecfield)
streams = []

Nq = 40
q = np.zeros((Nq, 3))
q[:, 1] = np.linspace(-1.5, 1.5, Nq)
q[:, 0] = 1.5
extent = np.array([-1.5, 1.5, -1.5, 1.5, 0, 0])
for qi in q:
    stream = mlab.pipeline.streamline(
        vecnorm,
        seed_scale=0.01,
        seedtype="point",
        integration_direction="both",
        extent=np.array([-1.5, 1.5, -1.5, 1.5, 0, 0]),
        colormap="viridis",
    )
    stream.stream_tracer.initial_integration_step = 0.1
    stream.stream_tracer.maximum_propagation = 200.0
    stream.seed.widget = stream.seed.widget_list[3]
    stream.seed.widget.position = qi
    stream.seed.widget.enabled = False  # hide the widget itself
    streams.append(stream)
